# For multiproc attach, we need to use a helper stub to import debuggee before running
# Flask; otherwise, we will get the connection only from the subprocess, not from the
# Flask server process.

import debuggee
import runpy

debuggee.setup()
runpy.run_module("flask", run_name="__main__", alter_sys=True)



import sys

if __name__ == '__main__':
    if 'in-sub' not in sys.argv:
        import os
        # These functions all execute a new program, replacing the current process; they do not return.
        # os.execl(path, arg0, arg1, ...)
        # os.execle(path, arg0, arg1, ..., env)
        # os.execlp(file, arg0, arg1, ...)
        # os.execlpe(file, arg0, arg1, ..., env)¶
        # os.execv(path, args)
        # os.execve(path, args, env)
        # os.execvp(file, args)
        # os.execvpe(file, args, env)
        os.execvp(sys.executable, [sys.executable, __file__, 'in-sub'])
    else:
        print('In sub')
        print('TEST SUCEEDED!')



import pydevd
import threading
import sys

original = pydevd.PyDB.notify_thread_created

found = set()


def new_notify_thread_created(self, thread_id, thread, *args, **kwargs):
    found.add(thread)
    return original(self, thread_id, thread, *args, **kwargs)


pydevd.PyDB.notify_thread_created = new_notify_thread_created

ok = []


class MyThread(threading.Thread):
    def run(self):
        if self not in found:
            ok.append(False)
        else:
            ok.append(True)


class ManualCreatedThreadPy313:
    def __init__(self):
        self.ev = threading.Event()

    def run(self):
        try:
            if threading.current_thread() not in found:
                ok.append(False)
            else:
                ok.append(True)
        finally:
            self.ev.set()

    def start(self):
        import _thread

        _thread.start_joinable_thread(self.run)

    def join(self):
        self.ev.wait()


class ManualCreatedThreadFromThreadModule:
    def __init__(self):
        self.ev = threading.Event()

    def run(self):
        try:
            if threading.current_thread() not in found:
                ok.append(False)
            else:
                ok.append(True)
        finally:
            self.ev.set()

    def start(self):
        try:
            import thread
        except Exception:
            import _thread as thread

        thread.start_new_thread(self.run)

    def join(self):
        self.ev.wait()


if __name__ == "__main__":
    threads: list = []

    if sys.version_info[:2] >= (3, 13):
        t1 = ManualCreatedThreadPy313()
        t1.start()
        threads.append(t1)

    t2 = ManualCreatedThreadFromThreadModule()
    t2.start()
    threads.append(t2)

    for i in range(15):
        t = MyThread()
        t.start()
        threads.append(t)

    for t in threads:
        t.join()

    assert len(ok) == len(threads)
    assert all(ok), "Expected all threads to be notified of their creation before starting to run. Found: %s" % (ok,)

    found.clear()
    print("TEST SUCEEDED")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

import tests
from tests import debug, test_data
from tests.debug import targets
from tests.patterns import some


if not tests.full:

    @pytest.fixture(params=targets.all_named)
    def target(request):
        return request.param


def test_with_dot_remote_root(pyfile, long_tmpdir, target, run):
    @pyfile
    def code_to_debug():
        import os
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(os.path.abspath(__file__))
        print("done")  # @bp

    dir_local = long_tmpdir.mkdir("local")
    dir_remote = long_tmpdir.mkdir("remote")

    path_local = dir_local / "code_to_debug.py"
    path_remote = dir_remote / "code_to_debug.py"

    code_to_debug.copy(path_local)
    code_to_debug.copy(path_remote)

    with debug.Session() as session:
        session.config["pathMappings"] = [{"localRoot": dir_local, "remoteRoot": "."}]

        backchannel = session.open_backchannel()
        with run(session, target(path_remote), cwd=dir_remote):
            # Set breakpoints using local path. This tests that local paths are
            # mapped to remote paths.
            session.set_breakpoints(path_local, all)

        actual_path_remote = backchannel.receive()
        assert some.path(actual_path_remote) == path_remote

        session.wait_for_stop(
            "breakpoint",
            expected_frames=[some.dap.frame(some.dap.source(path_local), line="bp")],
        )

        session.request_continue()


def test_with_path_mappings(pyfile, long_tmpdir, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import os
        import sys
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(os.path.abspath(__file__))
        call_me_back_dir = backchannel.receive()
        sys.path.insert(0, call_me_back_dir)

        import call_me_back

        def call_func():
            print("break here")  # @bp

        call_me_back.call_me_back(call_func)  # @call_me_back
        print("done")

    dir_local = long_tmpdir.mkdir("local")
    dir_remote = long_tmpdir.mkdir("remote")

    path_local = dir_local / "code_to_debug.py"
    path_remote = dir_remote / "code_to_debug.py"

    code_to_debug.copy(path_local)
    code_to_debug.copy(path_remote)

    call_me_back_dir = test_data / "call_me_back"
    call_me_back_py = call_me_back_dir / "call_me_back.py"

    with debug.Session() as session:
        session.config["pathMappings"] = [
            {"localRoot": dir_local, "remoteRoot": dir_remote}
        ]

        backchannel = session.open_backchannel()
        with run(session, target(path_remote)):
            # Set breakpoints using local path. This tests that local paths are
            # mapped to remote paths.
            session.set_breakpoints(path_local, ["bp"])

        actual_path_remote = backchannel.receive()
        assert some.path(actual_path_remote) == path_remote
        backchannel.send(call_me_back_dir)

        stop = session.wait_for_stop(
            "breakpoint",
            expected_frames=[
                some.dap.frame(
                    # Mapped files should not have a sourceReference, so that the IDE
                    # doesn't try to fetch them instead of opening the local file.
                    some.dap.source(path_local, sourceReference=0),
                    line="bp",
                ),
                some.dap.frame(
                    # Unmapped files should have a sourceReference, since there's no
                    # local file for the IDE to open.
                    some.dap.source(
                        call_me_back_py, sourceReference=some.int.not_equal_to(0)
                    ),
                    line="callback",
                ),
                some.dap.frame(
                    # Mapped files should not have a sourceReference, so that the IDE
                    # doesn't try to fetch them instead of opening the local file.
                    some.dap.source(path_local, sourceReference=0),
                    line="call_me_back",
                ),
            ],
        )

        srcref = stop.frames[1]["source"]["sourceReference"]

        try:
            session.request("source", {"sourceReference": 0})
        except Exception as exc:
            assert "Source unavailable" in str(exc)
        else:
            pytest.fail("sourceReference=0 should not be valid")

        source = session.request("source", {"sourceReference": srcref})
        assert "def call_me_back(callback):" in source["content"]

        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.


def test_vendoring(pyfile):
    @pyfile
    def import_debugpy():
        import sys

        class Dummy(object):
            __file__ = None

        class Dummy2(object):
            pass

        class Dummy3(object):
            def __getattribute__(self, *args, **kwargs):
                raise AssertionError("Error")

        sys.modules["pydev_dummy"] = Dummy()
        sys.modules["pydev_dummy2"] = Dummy2()
        sys.modules["pydev_dummy3"] = Dummy3()

        sys.modules["_pydev_dummy"] = Dummy()
        sys.modules["_pydev_dummy2"] = Dummy2()
        sys.modules["_pydev_dummy3"] = Dummy3()

        from debugpy._vendored import force_pydevd  # noqa

        print("Properly applied pydevd vendoring.")

    import subprocess
    import sys

    output = subprocess.check_output([sys.executable, "-u", import_debugpy.strpath])
    output = output.decode("utf-8")
    if "Properly applied pydevd vendoring." not in output:
        raise AssertionError(output)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from debugpy.common import messaging
from tests import debug


expected_at_line = {
    "in_do_something": [
        {"label": "SomeClass", "type": "class", "start": 0, "length": 4},
        {"label": "someFunction", "type": "function", "start": 0, "length": 4},
        {"label": "someVariable", "type": "field", "start": 0, "length": 4},
    ],
    "in_some_function": [
        {"label": "SomeClass", "type": "class", "start": 0, "length": 4},
        {"label": "someFunction", "type": "function", "start": 0, "length": 4},
        {"label": "someVar", "type": "field", "start": 0, "length": 4},
        {"label": "someVariable", "type": "field", "start": 0, "length": 4},
    ],
    "done": [
        {"label": "SomeClass", "type": "class", "start": 0, "length": 4},
        {"label": "someFunction", "type": "function", "start": 0, "length": 4},
    ],
}


@pytest.mark.parametrize("line", sorted(expected_at_line.keys()))
def test_completions_scope(pyfile, line, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        class SomeClass:
            def __init__(self, someVar):
                self.some_var = someVar

            def do_someting(self):
                someVariable = self.some_var
                return someVariable  # @in_do_something

        def someFunction(someVar):
            someVariable = someVar
            return SomeClass(someVariable).do_someting()  # @in_some_function

        someFunction("value")
        print("done")  # @done

    expected = sorted(expected_at_line[line], key=lambda t: t["label"])

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, [code_to_debug.lines[line]])

        stop = session.wait_for_stop("breakpoint")
        completions = session.request(
            "completions", {"text": "some", "frameId": stop.frame_id, "column": 5}
        )
        targets = sorted(completions["targets"], key=lambda t: t["label"])
        assert targets == expected

        session.request_continue()


def test_completions_cases(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()
        a = 1
        b = {"one": 1, "two": 2}
        c = 3
        print([a, b, c])  # @break

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, [code_to_debug.lines["break"]])

        stop = session.wait_for_stop()

        completions = session.request(
            "completions", {"frameId": stop.frame_id, "text": "b.", "column": 3}
        )
        labels = set(target["label"] for target in completions["targets"])
        assert labels >= {"get", "items", "keys", "setdefault", "update", "values"}

        completions = session.request(
            "completions",
            {"frameId": stop.frame_id, "text": "x = b.setdefault", "column": 13},
        )
        assert completions["targets"] == [
            {"label": "setdefault", "length": 6, "start": 6, "type": "function"}
        ]

        completions = session.request(
            "completions", {"frameId": stop.frame_id, "text": "not_there", "column": 10}
        )
        assert not completions["targets"]

        with pytest.raises(messaging.MessageHandlingError):
            completions = session.request(
                "completions",
                {
                    "frameId": 9999999,  # nonexistent frameId
                    "text": "not_there",
                    "column": 10,
                },
            )

        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

from tests import debug
from tests.patterns import some


def test_tracing(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy

        debuggee.setup()

        def func(expected_tracing):
            print(1)  # @inner1

            # Test nested change/restore. Going from False to True only works entirely
            # correctly on Python 3.6+; on earlier versions, if tracing wasn't enabled
            # when the function is entered, re-enabling it later will not cause the
            # breakpoints in this function to light up. However, it will allow hitting
            # breakpoints in functions called from here.

            def inner2():
                print(2)  # @inner2

            debugpy.trace_this_thread(not expected_tracing)
            inner2()
            debugpy.trace_this_thread(expected_tracing)

            print(3)  # @inner3

        debugpy.trace_this_thread(False)
        print(0)  # @outer1
        func(False)

        debugpy.trace_this_thread(True)
        print(0)  # @outer2
        func(True)

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop(expected_frames=[some.dap.frame(code_to_debug, "inner2")])
        session.request_continue()

        session.wait_for_stop(expected_frames=[some.dap.frame(code_to_debug, "outer2")])
        session.request_continue()

        session.wait_for_stop(expected_frames=[some.dap.frame(code_to_debug, "inner1")])
        session.request_continue()

        session.wait_for_stop(expected_frames=[some.dap.frame(code_to_debug, "inner3")])
        session.request_continue()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""debugpy tests
"""

import os
import pkgutil
import py
import pytest

# Do not import anything from debugpy until assert rewriting is enabled below!

full = int(os.environ.get("DEBUGPY_TESTS_FULL", "0")) != 0

root = py.path.local(__file__) / ".."

test_data = root / "test_data"
"""A py.path.local object for the tests/test_data/ directory.

Idiomatic use is via from .. import::

    from tests import test_data
    f = open(str(test_data / "attach" / "attach1.py"))
"""


# This is only imported to ensure that the module is actually installed and the
# timeout setting in pytest.ini is active, since otherwise most timeline-based
# tests will hang indefinitely if they time out.
import pytest_timeout  # noqa


# We want pytest to rewrite asserts (for better error messages) in the common code
# code used by the tests, and in all the test helpers. This does not affect debugpy
# inside debugged processes.


def _register_assert_rewrite(modname):
    modname = str(modname)
    # print("pytest.register_assert_rewrite({0!r})".format(modname))
    pytest.register_assert_rewrite(modname)


_register_assert_rewrite("debugpy.common")
tests_submodules = pkgutil.iter_modules([str(root)])
for _, submodule, _ in tests_submodules:
    submodule = str("{0}.{1}".format(__name__, submodule))
    _register_assert_rewrite(submodule)


# Now we can import these, and pytest will rewrite asserts in them.
from debugpy.common import json, log
import debugpy.server  # noqa

# Clean up environment variables that were automatically set when importing pydevd -
# we don't need them in the test runner process (since pydevd is not tracing it),
# and some tests must be able to spawn debuggee with them unset.
for name in (
    "DEBUGPY_LOG_DIR",
    "PYDEVD_DEBUG",
    "PYDEVD_DEBUG_FILE",
    "PYDEVD_USE_FRAME_EVAL",
):
    os.environ.pop(name, None)

# Enable full logging to stderr, and make timestamps shorter to match maximum test
# run time better.
log.stderr.levels = all
log.timestamp_format = "06.3f"
log.to_file(prefix="tests")


# Enable JSON serialization for py.path.local.
def json_default(self, obj):
    if isinstance(obj, py.path.local):
        return obj.strpath
    return self.original_default(obj)


json.JsonEncoder.original_default = json.JsonEncoder.default
json.JsonEncoder.default = json_default



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Pattern matching for recursive Python data structures.

Usage::

    from tests.patterns import some

    assert object() == some.object
    assert None == some.object
    assert None != some.thing
    assert None == ~some.thing  # inverse
    assert None == some.thing | None

    assert 123 == some.thing.in_range(0, 200)
    assert "abc" == some.thing.such_that(lambda s: s.startswith("ab"))

    xs = []
    assert xs == some.specific_object(xs)  # xs is xs
    assert xs != some.specific_object([])  # xs is not []

    assert Exception() == some.instanceof(Exception)
    assert 123 == some.instanceof((int, str))
    assert "abc" == some.instanceof((int, str))

    assert True == some.bool
    assert 123.456 == some.number
    assert 123 == some.int
    assert Exception() == some.error
    assert object() == some.object.same_as(object())

    assert b"abc" == some.bytes
    assert "abc" == some.str
    assert b"abc" != some.str

    assert "abbc" == some.str.starting_with("ab")
    assert "abbc" == some.str.ending_with("bc")
    assert "abbc" == some.str.containing("bb")

    assert "abbc" == some.str.matching(r".(b+).")
    assert "abbc" != some.str.matching(r"ab")
    assert "abbc" != some.str.matching(r"bc")

    if sys.platform == "win32":
        assert "\\Foo\\Bar" == some.path("/foo/bar")
    else:
        assert "/Foo/Bar" != some.path("/foo/bar")

    assert {
        "bool": True,
        "list": [None, True, 123],
        "dict": {
            "int": 123,
            "str": "abc",
        },
    } == some.dict.containing({
        "list": [None, some.bool, some.int | some.str],
        "dict": some.dict.containing({
            "int": some.int.in_range(100, 200),
        })
    })
"""

__all__ = [
    "bool",
    "bytes",
    "dap",
    "dict",
    "error",
    "instanceof",
    "int",
    "list",
    "number",
    "path",
    "str",
    "thing",
    "tuple",
]

import builtins
import numbers
import re

from tests.patterns import _impl


object = _impl.Object()
thing = _impl.Thing()
instanceof = _impl.InstanceOf
path = _impl.Path


bool = instanceof(builtins.bool)
number = instanceof(numbers.Real, "number")
int = instanceof(numbers.Integral, "int")
tuple = instanceof(builtins.tuple)
error = instanceof(Exception)


bytes = instanceof(builtins.bytes)
bytes.starting_with = lambda prefix: bytes.matching(
    re.escape(prefix) + b".*", re.DOTALL
)
bytes.ending_with = lambda suffix: bytes.matching(b".*" + re.escape(suffix), re.DOTALL)
bytes.containing = lambda sub: bytes.matching(b".*" + re.escape(sub) + b".*", re.DOTALL)


str = instanceof(builtins.str)
str.starting_with = lambda prefix: str.matching(re.escape(prefix) + ".*", re.DOTALL)
str.ending_with = lambda suffix: str.matching(".*" + re.escape(suffix), re.DOTALL)
str.containing = lambda sub: str.matching(".*" + re.escape(sub) + ".*", re.DOTALL)


list = instanceof(builtins.list)
list.containing = _impl.ListContaining


dict = instanceof(builtins.dict)
dict.containing = _impl.DictContaining


# Set in __init__.py to avoid circular dependency.
dap = None



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""The main script for the watchdog worker process.
"""

# This code runs in a separate process, and should not import pytest or tests!
# Do not import debugpy on top level, either - sys.path needs to be fixed first -
# this is done in main().

import collections
import psutil
import sys
import time


ProcessInfo = collections.namedtuple("ProcessInfo", ["process", "name"])


def main(tests_pid):
    from debugpy.common import log, messaging

    # log.stderr_levels |= {"info"}
    log.timestamp_format = "06.3f"
    log_file = log.to_file(prefix="tests.watchdog")

    stream = messaging.JsonIOStream.from_stdio(f"tests-{tests_pid}")
    log.info("Spawned WatchDog-{0} for tests-{0}", tests_pid)
    tests_process = psutil.Process(tests_pid)
    stream.write_json(["watchdog", log_file.filename])

    spawned_processes = {}  # pid -> ProcessInfo
    try:
        stop = False
        while not stop:
            try:
                message = stream.read_json()
            except Exception:
                break

            command = message[0]
            args = message[1:]

            if command == "stop":
                assert not args
                stop = True

            elif command == "register_spawn":
                pid, name = args
                pid = int(pid)

                log.info(
                    "WatchDog-{0} registering spawned process {1} (pid={2})",
                    tests_pid,
                    name,
                    pid,
                )
                try:
                    _, old_name = spawned_processes[pid]
                except KeyError:
                    pass
                else:
                    log.warning(
                        "WatchDog-{0} already tracks a process with pid={1}: {2}",
                        tests_pid,
                        pid,
                        old_name,
                    )
                spawned_processes[pid] = ProcessInfo(psutil.Process(pid), name)

            elif command == "unregister_spawn":
                pid, name = args
                pid = int(pid)

                log.info(
                    "WatchDog-{0} unregistering spawned process {1} (pid={2})",
                    tests_pid,
                    name,
                    pid,
                )
                spawned_processes.pop(pid, None)

            else:
                raise AssertionError(f"Unknown watchdog command: {command!r}")

            stream.write_json(["ok"])

    except Exception as exc:
        stream.write_json(["error", str(exc)])
        log.reraise_exception()

    finally:
        try:
            stream.close()
        except Exception:
            log.swallow_exception()

        # If the test runner becomes a zombie process, it is still considered alive,
        # and wait() will block indefinitely. Poll status instead.
        while True:
            try:
                status = tests_process.status()
            except Exception:
                # If we can't even get its status, assume that it's dead.
                break

            # If it's dead or a zombie, time to clean it up.
            if status in (psutil.STATUS_DEAD, psutil.STATUS_ZOMBIE):
                break

            # Otherwise, let's wait a bit to see if anything changes.
            try:
                tests_process.wait(0.1)
            except Exception:
                pass

        leftover_processes = {proc for proc, _ in spawned_processes.values()}
        for proc, _ in spawned_processes.values():
            try:
                leftover_processes |= proc.children(recursive=True)
            except Exception:
                pass

        leftover_processes = {proc for proc in leftover_processes if proc.is_running()}
        if not leftover_processes:
            return

        # Wait a bit to allow the terminal to catch up on the test runner output.
        time.sleep(0.3)

        log.newline(level="warning")
        log.warning(
            "tests-{0} process terminated unexpectedly, and left some orphan child "
            "processes behind: {1!r}",
            tests_pid,
            sorted({proc.pid for proc in leftover_processes}),
        )

        for proc in leftover_processes:
            log.warning(
                "WatchDog-{0} killing orphaned test child process (pid={1})",
                tests_pid,
                proc.pid,
            )

            try:
                proc.kill()
            except psutil.NoSuchProcess:
                pass
            except Exception:
                log.swallow_exception()

        log.info("WatchDog-{0} exiting", tests_pid)


if __name__ == "__main__":
    main(int(sys.argv[1]))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Test helpers for networking.
"""

import os
import re
import requests
import socket
import threading
import time

from debugpy.common import log, util
from tests.patterns import some

used_ports = set()
    
def get_test_server_port():
    """Returns a server port number that can be safely used for listening without
    clashing with another test worker process, when running with pytest-xdist.

    If multiple test workers invoke this function with the same min value, each of
    them will receive a different number that is not lower than start (but may be
    higher). If the resulting value is >=stop, it is a fatal error.

    Note that if multiple test workers invoke this function with different ranges
    that overlap, conflicts are possible!
    """

    try:
        worker_id = util.force_ascii(os.environ["PYTEST_XDIST_WORKER"])
    except KeyError:
        n = 0
    else:
        assert worker_id == some.bytes.matching(
            rb"gw(\d+)"
        ), "Unrecognized PYTEST_XDIST_WORKER format"
        n = int(worker_id[2:])

    port = 5678 + (n * 300)
    while port in used_ports:
        port += 1
    used_ports.add(port)

    return port


def find_http_url(text):
    match = re.search(r"https?://[-.0-9A-Za-z]+(:\d+)/?", text)
    return match.group() if match else None


def wait_until_port_is_listening(port, interval=1, max_attempts=1000):
    """Blocks until the specified TCP port on localhost is listening, and can be
    connected to.

    Tries to connect to the port periodically, and repeats until connection succeeds.
    Connection is immediately closed before returning.
    """

    for i in range(1, max_attempts + 1):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            log.info("Probing localhost:{0} (attempt {1})...", port, i)
            sock.connect(("localhost", port))
        except socket.error as exc:
            # The first attempt will almost always fail, because the port isn't
            # open yet. But if it keeps failing after that, we want to know why.
            if i > 1:
                log.warning("Failed to connect to localhost:{0}:\n{1}", port, exc)
            time.sleep(interval)
        else:
            log.info("localhost:{0} is listening - server is up!", port)
            return
        finally:
            sock.close()


class WebRequest(object):
    """An async wrapper around requests."""

    @staticmethod
    def get(*args, **kwargs):
        return WebRequest("get", *args, **kwargs)

    @staticmethod
    def post(*args, **kwargs):
        return WebRequest("post", *args, **kwargs)

    def __init__(self, method, url, *args, **kwargs):
        """Invokes requests.method(url, *args, **kwargs) on a background thread,
        and immediately returns.

        If method() raises an exception, it is logged, unless log_errors=False.
        """

        self.method = method
        self.url = url

        self.log_errors = kwargs.pop("log_errors", True)

        self.request = None
        """The underlying requests.Request object.

        Not set until wait_for_response() returns.
        """

        self.exception = None
        """Exception that occurred while performing the request, if any.

        Not set until wait_for_response() returns.
        """

        log.info("{0}", self)

        func = getattr(requests, method)
        self._worker_thread = threading.Thread(
            target=lambda: self._worker(func, *args, **kwargs),
            name=f"WebRequest({self})",
        )
        self._worker_thread.daemon = True
        self._worker_thread.start()

    def __str__(self):
        return f"HTTP {self.method.upper()} {self.url}"

    def _worker(self, func, *args, **kwargs):
        try:
            self.request = func(self.url, *args, **kwargs)
        except Exception as exc:
            if self.log_errors:
                log.swallow_exception("{0} failed:", self)
            self.exception = exc
        else:
            log.info(
                "{0} --> {1} {2}", self, self.request.status_code, self.request.reason
            )

    def wait_for_response(self, timeout=None):
        """Blocks until the request completes, and returns self.request."""
        if self._worker_thread.is_alive():
            log.info("Waiting for response to {0} ...", self)
            self._worker_thread.join(timeout)

        if self.exception is not None:
            raise self.exception
        return self.request

    def response_text(self):
        """Blocks until the request completes, and returns the response body."""
        return self.wait_for_response().text


class WebServer(object):
    """Interacts with a web server listening on localhost on the specified port."""

    def __init__(self, port):
        self.port = port
        self.url = f"http://localhost:{port}"

    def __enter__(self):
        """Blocks until the server starts listening on self.port."""
        log.info("Web server expected on {0}", self.url)
        wait_until_port_is_listening(self.port, interval=3)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Sends an HTTP /exit GET request to the server.

        The server is expected to terminate its process while handling that request.
        """
        self.get("/exit", log_errors=False)

    def get(self, path, *args, **kwargs):
        return WebRequest.get(self.url + path, *args, **kwargs)

    def post(self, path, *args, **kwargs):
        return WebRequest.post(self.url + path, *args, **kwargs)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

# Expose Session directly.
from tests.debug.session import Session  # noqa



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

import debugpy
from debugpy.common import log
from tests import debug
from tests.patterns import some


@pytest.fixture
def expected_system_info():
    def version_str(v):
        return "%d.%d.%d%s%d" % (v.major, v.minor, v.micro, v.releaselevel, v.serial)

    try:
        impl_name = sys.implementation.name
    except AttributeError:
        impl_name = ""

    try:
        impl_version = version_str(sys.implementation.version)
    except AttributeError:
        impl_version = ""

    return some.dict.containing(
        {
            "debugpy": some.dict.containing({"version": debugpy.__version__}),
            "python": some.dict.containing(
                {
                    "version": version_str(sys.version_info),
                    "implementation": some.dict.containing(
                        {
                            "name": impl_name,
                            "version": impl_version,
                            "description": some.str,
                        }
                    ),
                }
            ),
            "platform": some.dict.containing({"name": sys.platform}),
            "process": some.dict.containing(
                {
                    "pid": some.int,
                    "executable": sys.executable,
                    "bitness": 64 if sys.maxsize > 2**32 else 32,
                }
            ),
        }
    )


def test_debugpySystemInfo(pyfile, target, run, expected_system_info):
    @pyfile
    def code_to_debug():
        import debuggee
        import debugpy

        debuggee.setup()
        debugpy.breakpoint()
        print()

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            pass

        session.wait_for_stop()

        system_info = session.request("debugpySystemInfo")
        log.info("Expected system info: {0}", expected_system_info)
        assert system_info == expected_system_info

        session.request_continue()



import threading
import unittest
import sys
import pydevconsole
from _pydev_bundle.pydev_imports import xmlrpclib, SimpleXMLRPCServer
from _pydevd_bundle import pydevd_io
from contextlib import contextmanager
import pytest

try:
    from ast import PyCF_ALLOW_TOP_LEVEL_AWAIT  # @UnusedImport

    CAN_EVALUATE_TOP_LEVEL_ASYNC = True
except:
    CAN_EVALUATE_TOP_LEVEL_ASYNC = False


# =======================================================================================================================
# Test
# =======================================================================================================================
class Test(unittest.TestCase):
    @contextmanager
    def interpreter(self):
        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr
        sys.stdout = pydevd_io.IOBuf()
        sys.stderr = pydevd_io.IOBuf()
        try:
            sys.stdout.encoding = sys.stdin.encoding
            sys.stderr.encoding = sys.stdin.encoding
        except AttributeError:
            # In Python 3 encoding is not writable (whereas in Python 2 it doesn't exist).
            pass

        try:
            client_port, _server_port = self.get_free_addresses()
            client_thread = self.start_client_thread(client_port)  # @UnusedVariable
            import time

            time.sleep(0.3)  # let's give it some time to start the threads

            from _pydev_bundle import pydev_localhost

            interpreter = pydevconsole.InterpreterInterface(pydev_localhost.get_localhost(), client_port, threading.current_thread())
            yield interpreter
        except:
            # if there's some error, print the output to the actual output.
            self.original_stdout.write(sys.stdout.getvalue())
            self.original_stderr.write(sys.stderr.getvalue())
            raise
        finally:
            sys.stderr = self.original_stderr
            sys.stdout = self.original_stdout

    def test_console_hello(self):
        with self.interpreter() as interpreter:
            (result,) = interpreter.hello("Hello pydevconsole")
            self.assertEqual(result, "Hello eclipse")

    @pytest.mark.skipif(not CAN_EVALUATE_TOP_LEVEL_ASYNC, reason="Requires top-level async.")
    def test_console_async(self):
        with self.interpreter() as interpreter:
            from _pydev_bundle.pydev_console_utils import CodeFragment

            more = interpreter.add_exec(
                CodeFragment(
                    """
async def async_func(a):
    return a
"""
                )
            )
            assert not more
            assert not sys.stderr.getvalue()
            assert not sys.stdout.getvalue()

            more = interpreter.add_exec(CodeFragment("""x = await async_func(1111)"""))
            assert not more
            assert not sys.stderr.getvalue()
            assert not sys.stdout.getvalue()

            more = interpreter.add_exec(CodeFragment("""print(x)"""))
            assert not more
            assert not sys.stderr.getvalue()
            assert "1111" in sys.stdout.getvalue()

    def test_console_requests(self):
        with self.interpreter() as interpreter:
            from _pydev_bundle.pydev_console_utils import CodeFragment

            interpreter.add_exec(CodeFragment("class Foo:\n    CONSTANT=1\n"))
            interpreter.add_exec(CodeFragment("foo=Foo()"))
            interpreter.add_exec(CodeFragment("foo.__doc__=None"))
            interpreter.add_exec(CodeFragment("val = input()"))
            interpreter.add_exec(CodeFragment("50"))
            interpreter.add_exec(CodeFragment("print (val)"))
            found = sys.stdout.getvalue().split()
            try:
                self.assertEqual(["50", "input_request"], found)
            except:
                try:
                    self.assertEqual(["input_request"], found)  # IPython
                except:
                    self.assertEqual(["50", "input_request"], found[1:])  # IPython 5.1
                    self.assertTrue(found[0].startswith("Out"))

            comps = interpreter.getCompletions("foo.", "foo.")
            self.assertTrue(("CONSTANT", "", "", "3") in comps or ("CONSTANT", "", "", "4") in comps, "Found: %s" % comps)

            comps = interpreter.getCompletions('"".', '"".')
            self.assertTrue(
                ("__add__", "x.__add__(y) <==> x+y", "", "3") in comps
                or ("__add__", "", "", "4") in comps
                or ("__add__", "x.__add__(y) <==> x+y\r\nx.__add__(y) <==> x+y", "()", "2") in comps
                or ("__add__", "x.\n__add__(y) <==> x+yx.\n__add__(y) <==> x+y", "()", "2"),
                "Did not find __add__ in : %s" % (comps,),
            )

            completions = interpreter.getCompletions("", "")
            for c in completions:
                if c[0] == "AssertionError":
                    break
            else:
                self.fail("Could not find AssertionError")

            completions = interpreter.getCompletions("Assert", "Assert")
            for c in completions:
                if c[0] == "RuntimeError":
                    self.fail("Did not expect to find RuntimeError there")

            assert ("__doc__", None, "", "3") not in interpreter.getCompletions("foo.CO", "foo.")

            comps = interpreter.getCompletions("va", "va")
            assert ("val", "", "", "3") in comps or ("val", "", "", "4") in comps

            interpreter.add_exec(CodeFragment('s = "mystring"'))

            desc = interpreter.getDescription("val")
            self.assertTrue(
                desc.find("str(object) -> string") >= 0
                or desc == "'input_request'"
                or desc.find("str(string[, encoding[, errors]]) -> str") >= 0
                or desc.find("str(Char* value)") >= 0
                or desc.find("str(object='') -> string") >= 0
                or desc.find("str(value: Char*)") >= 0
                or desc.find("str(object='') -> str") >= 0
                or desc.find("The most base type") >= 0,  # Jython 2.7 is providing this :P
                "Could not find what was needed in %s" % desc,
            )

            desc = interpreter.getDescription("val.join")
            self.assertTrue(
                desc.find("S.join(sequence) -> string") >= 0
                or desc.find("S.join(sequence) -> str") >= 0
                or desc.find("S.join(iterable) -> string") >= 0
                or desc == "<builtin method 'join'>"
                or desc == "<built-in method join of str object>"
                or desc.find("str join(str self, list sequence)") >= 0
                or desc.find("S.join(iterable) -> str") >= 0
                or desc.find("join(self: str, sequence: list) -> str") >= 0
                or desc.find("Concatenate any number of strings.") >= 0
                or desc.find("bound method str.join") >= 0,  # PyPy
                "Could not recognize: %s" % (desc,),
            )

    def start_client_thread(self, client_port):
        class ClientThread(threading.Thread):
            def __init__(self, client_port):
                threading.Thread.__init__(self)
                self.client_port = client_port

            def run(self):
                class HandleRequestInput:
                    def RequestInput(self):
                        client_thread.requested_input = True
                        return "input_request"

                    def NotifyFinished(self, *args, **kwargs):
                        client_thread.notified_finished += 1
                        return 1

                handle_request_input = HandleRequestInput()

                from _pydev_bundle import pydev_localhost

                client_server = SimpleXMLRPCServer((pydev_localhost.get_localhost(), self.client_port), logRequests=False)
                client_server.register_function(handle_request_input.RequestInput)
                client_server.register_function(handle_request_input.NotifyFinished)
                client_server.serve_forever()

        client_thread = ClientThread(client_port)
        client_thread.requested_input = False
        client_thread.notified_finished = 0
        client_thread.daemon = True
        client_thread.start()
        return client_thread

    def start_debugger_server_thread(self, debugger_port, socket_code):
        class DebuggerServerThread(threading.Thread):
            def __init__(self, debugger_port, socket_code):
                threading.Thread.__init__(self)
                self.debugger_port = debugger_port
                self.socket_code = socket_code

            def run(self):
                import socket

                s = socket.socket()
                s.bind(("", debugger_port))
                s.listen(1)
                socket, unused_addr = s.accept()
                socket_code(socket)

        debugger_thread = DebuggerServerThread(debugger_port, socket_code)
        debugger_thread.daemon = True
        debugger_thread.start()
        return debugger_thread

    def get_free_addresses(self):
        from _pydev_bundle.pydev_localhost import get_socket_names

        socket_names = get_socket_names(2, True)
        port0 = socket_names[0][1]
        port1 = socket_names[1][1]

        assert port0 != port1
        assert port0 > 0
        assert port1 > 0

        return port0, port1

    def test_server(self):
        self.original_stdout = sys.stdout
        sys.stdout = pydevd_io.IOBuf()
        try:
            client_port, server_port = self.get_free_addresses()

            class ServerThread(threading.Thread):
                def __init__(self, client_port, server_port):
                    threading.Thread.__init__(self)
                    self.client_port = client_port
                    self.server_port = server_port

                def run(self):
                    from _pydev_bundle import pydev_localhost

                    pydevconsole.start_server(pydev_localhost.get_localhost(), self.server_port, self.client_port)

            server_thread = ServerThread(client_port, server_port)
            server_thread.daemon = True
            server_thread.start()

            client_thread = self.start_client_thread(client_port)  # @UnusedVariable

            import time

            time.sleep(0.3)  # let's give it some time to start the threads
            sys.stdout = pydevd_io.IOBuf()

            from _pydev_bundle import pydev_localhost

            server = xmlrpclib.Server("http://%s:%s" % (pydev_localhost.get_localhost(), server_port))
            server.execLine("class Foo:")
            server.execLine("    pass")
            server.execLine("")
            server.execLine("foo = Foo()")
            server.execLine("a = input()")
            server.execLine("print (a)")
            initial = time.time()
            while not client_thread.requested_input:
                if time.time() - initial > 2:
                    raise AssertionError("Did not get the return asked before the timeout.")
                time.sleep(0.1)

            found = sys.stdout.getvalue()
            while ["input_request"] != found.split():
                found += sys.stdout.getvalue()
                if time.time() - initial > 2:
                    break
                time.sleep(0.1)
            self.assertEqual(["input_request"], found.split())
        finally:
            sys.stdout = self.original_stdout



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""pytest configuration.
"""

pytest_plugins = ["tests.pytest_fixtures", "tests.pytest_hooks"]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Patterns that are specific to the Debug Adapter Protocol.
"""

import py.path

from tests import code
from tests.patterns import some, _impl


id = some.int.in_range(0, 10000)
"""Matches a DAP "id", assuming some reasonable range for an implementation that
generates those ids sequentially.
"""


def source(path, **kwargs):
    """Matches DAP Source objects."""
    if isinstance(path, py.path.local):
        path = some.path(path)
    d = {"path": path}
    d.update(kwargs)
    return some.dict.containing(d)


def frame(source, line, column=some.int, **kwargs):
    """Matches DAP Frame objects.

    If source is py.path.local, it's automatically wrapped with some.dap.source().

    If line is str, it is treated as a line marker, and translated to a line
    number via get_marked_line_numbers(source["path"]) if possible.
    """

    if isinstance(source, py.path.local):
        source = some.dap.source(source)

    if isinstance(line, str):
        if isinstance(source, dict):
            path = source["path"]
        elif isinstance(source, _impl.DictContaining):
            path = source.items["path"]
        else:
            path = None
        assert isinstance(
            path, _impl.Path
        ), "source must be some.dap.source() to use line markers in some.dap.frame()"
        line = code.get_marked_line_numbers(path.path)[line]

    d = {"id": some.dap.id, "source": source, "line": line, "column": column}
    d.update(kwargs)
    return some.dict.containing(d)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import collections
import itertools
import os
import psutil
import py
import subprocess
import sys
import time

import debugpy.adapter
from debugpy.common import json, log, messaging, sockets, util
import tests
from tests import code, timeline, watchdog
from tests.debug import comms, config, output
from tests.patterns import some

DEBUGGEE_PYTHONPATH = tests.root / "DEBUGGEE_PYTHONPATH"

StopInfo = collections.namedtuple(
    "StopInfo", ["body", "frames", "thread_id", "frame_id"]
)


class Session(object):
    """A test debug session. Manages the lifetime of the adapter and the debuggee
    processes, captures debuggee stdio output, establishes a DAP message channel to
    the debuggee, and records all DAP messages in that channel on a Timeline object.

    Must be used in a with-statement for proper cleanup. On successful exit - if no
    exception escapes from the with-statement - the session will:

    1. Invoke wait_for_exit(), unless expected_exit_code is None.
    2. Invoke disconnect().
    3. Wait for the adapter process to exit.
    4. Finalize and closes the timeline

    If the exit is due to an exception, the session will:

    1. Invoke disconnect(force=True).
    2. Kill the debuggee and the adapter processes.

    Example::

        with debug.Session() as session:
            # Neither debuggee nor adapter are spawned yet. Initial configuration.
            session.log_dir = ...
            session.config.update({...})

            with session.launch(...):
                # Debuggee and adapter are spawned, but there is no code executing
                # in the debuggee yet.
                session.set_breakpoints(...)

            # Code is executing in the debuggee.
            session.wait_for_stop(expected_frames=[...])
            assert session.get_variable(...) == ...
            session.request_continue()

        # Session is disconnected from the debuggee, and both the debuggee and the
        # adapter processes have exited.
        assert session.exit_code == ...
    """

    tmpdir = None
    """Temporary directory in which Sessions can create the temp files they need.

    Automatically set to tmpdir for the current test by pytest_fixtures.test_wrapper().
    """

    @classmethod
    def reset_counter(cls):
        cls._counter = itertools.count(1)

    def __init__(self, debug_config=None):
        assert Session.tmpdir is not None
        watchdog.start()

        self.id = next(Session._counter)
        log.info("Starting {0}", self)

        self.client_id = "vscode"

        self.capabilities = {
            "pathFormat": "path",
            "clientID": self.client_id,
            "adapterID": "test",
            "linesStartAt1": True,
            "columnsStartAt1": True,
            "supportsVariableType": True,
            "supportsRunInTerminalRequest": True,
            "supportsArgsCanBeInterpretedByShell": True,
            "supportsStartDebuggingRequest": False,
        }

        self.debuggee = None
        """psutil.Popen instance for the debuggee process."""

        self.adapter = None
        """psutil.Popen instance for the adapter process."""

        self.expected_adapter_sockets = {
            "client": {"host": some.str, "port": some.int, "internal": False},
        }
        """The sockets which the adapter is expected to report."""

        self.adapter_endpoints = None
        """Name of the file that contains the adapter endpoints information.

        This file is generated by the adapter when it opens the listener sockets,
        and deleted by it when it exits.
        """

        self.channel = None
        """JsonMessageChannel to the adapter."""

        self.captured_output = {"stdout", "stderr"}
        """Before the debuggee is spawned, this is the set of stdio streams that
        should be captured once it is spawned.

        After it is spawned, this is a CapturedOutput object capturing those streams.
        """

        self.backchannel = None
        """The BackChannel object to talk to the debuggee.

        Must be explicitly created with open_backchannel().
        """

        self.scratchpad = comms.ScratchPad(self)
        """The ScratchPad object to talk to the debuggee."""

        self.start_command = None
        """Set to either "launch" or "attach" just before the corresponding request is sent.
        """

        self.start_request = None
        """The "launch" or "attach" request that started executing code in this session.
        """

        self.expected_exit_code = 0
        """The expected exit code for the debuggee process.

        If None, the debuggee is not expected to exit when the Session is closed.

        If not None, this is validated against both exit_code and debuggee.returncode.
        """

        self.exit_code = None
        """The actual exit code for the debuggee process, as received from DAP.
        """

        self.config = config.DebugConfig(
            debug_config
            if debug_config is not None
            else {
                "justMyCode": True,
                "name": "Test",
                "type": "python",
            }
        )
        """The debug configuration for this session."""

        self.before_connect = lambda address: None
        """Invoked right before a socket connection to the adapter is established.
        """

        self.before_request = lambda command, arguments: None
        """Invoked for every outgoing request in this session, allowing any final
        tweaks to the request before it is sent.
        """

        self.log_dir = (
            None if log.log_dir is None else py.path.local(log.log_dir) / str(self)
        )
        """The log directory for this session. Passed via DEBUGPY_LOG_DIR to all spawned
        child processes.

        If set to None, DEBUGPY_LOG_DIR is not automatically added, but tests can still
        provide it manually.
        """

        self.tmpdir = Session.tmpdir / str(self)
        self.tmpdir.ensure(dir=True)

        self.timeline = timeline.Timeline(str(self))
        self.ignore_unobserved.extend(
            [
                timeline.Event("module"),
                timeline.Event("continued"),
                timeline.Event("debugpyWaitingForServer"),
                timeline.Event("debugpySockets"),
                timeline.Event("thread", some.dict.containing({"reason": "started"})),
                timeline.Event("thread", some.dict.containing({"reason": "exited"})),
                timeline.Event("output", some.dict.containing({"category": "stdout"})),
                timeline.Event("output", some.dict.containing({"category": "stderr"})),
                timeline.Event("output", some.dict.containing({"category": "console"})),
                timeline.Event(
                    "output", some.dict.containing({"category": "important"})
                ),
            ]
        )

        # Expose some common members of timeline directly - these should be the ones
        # that are the most straightforward to use, and are difficult to use incorrectly.
        # Conversely, most tests should restrict themselves to this subset of the API,
        # and avoid calling members of timeline directly unless there is a good reason.
        self.new = self.timeline.new
        self.observe = self.timeline.observe
        self.wait_for_next = self.timeline.wait_for_next
        self.proceed = self.timeline.proceed
        self.expect_new = self.timeline.expect_new
        self.expect_realized = self.timeline.expect_realized
        self.all_occurrences_of = self.timeline.all_occurrences_of
        self.observe_all = self.timeline.observe_all

        spawn_adapter = self.spawn_adapter
        self.spawn_adapter = lambda *args, **kwargs: spawn_adapter(*args, **kwargs)
        self.spawn_adapter.env = util.Env()

        spawn_debuggee = self.spawn_debuggee
        self.spawn_debuggee = lambda *args, **kwargs: spawn_debuggee(*args, **kwargs)
        self.spawn_debuggee.env = util.Env()

    def __str__(self):
        return f"Session[{self.id}]"

    @property
    def adapter_id(self):
        return f"Adapter[{self.id}]"

    @property
    def debuggee_id(self):
        return f"Debuggee[{self.id}]"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        log.info("Ending {0}.", self)

        if self.timeline.is_frozen:
            self.timeline.unfreeze()

        # Only wait for exit if there was no exception in the test - if there was one,
        # the debuggee might still be waiting for further requests.
        if exc_type is None:
            # If expected_exit_code is set to None, the debuggee is not expected to
            # exit after this Session is closed (e.g. because another Session will
            # attach to it later on).
            if self.expected_exit_code is not None:
                self.wait_for_exit()
        else:
            # Log the error, in case another one happens during shutdown.
            log.swallow_exception(exc_info=(exc_type, exc_val, exc_tb))

        if exc_type is None:
            self.disconnect()
            self.timeline.close()
        else:
            # If there was an exception, don't try to send any more messages to avoid
            # spamming log with irrelevant entries - just close the channel and kill
            # all the processes immediately. Don't close or finalize the timeline,
            # either, since it'll likely have unobserved events in it.
            if self.adapter is not None:
                log.info("Killing {0}.", self.adapter_id)
                try:
                    self.adapter.kill()
                except Exception:
                    pass
            if self.debuggee is not None:
                log.info("Killing {0}.", self.debuggee_id)
                try:
                    self.debuggee.kill()
                except Exception:
                    pass
            self.disconnect(force=True)

        if self.adapter_endpoints is not None and self.expected_exit_code is not None:
            log.info("Waiting for {0} to close listener ports ...", self.adapter_id)
            while self.adapter_endpoints.check():
                time.sleep(0.1)

        if self.adapter is not None:
            log.info(
                "Waiting for {0} with PID={1} to exit.",
                self.adapter_id,
                self.adapter.pid,
            )
            self.adapter.wait()
            watchdog.unregister_spawn(self.adapter.pid, self.adapter_id)
            self.adapter = None

        if self.backchannel is not None:
            self.backchannel.close()
            self.backchannel = None

        # Work around https://bugs.python.org/issue37380
        for popen in self.debuggee, self.adapter:
            if popen is not None and popen.returncode is None:
                popen.returncode = -1

    @property
    def ignore_unobserved(self):
        return self.timeline.ignore_unobserved
    
    @property
    def is_subprocess(self):
        return "subProcessId" in self.config

    def open_backchannel(self):
        assert self.backchannel is None
        self.backchannel = comms.BackChannel(self)
        self.backchannel.listen()
        return self.backchannel

    def _init_log_dir(self):
        if self.log_dir is None:
            return False

        log.info("Logs for {0} will be in {1}", self, json.repr(self.log_dir))
        try:
            self.log_dir.remove()
        except Exception:
            pass
        self.log_dir.ensure(dir=True)

        # Make subsequent calls of this method no-op for the remainder of the session.
        self._init_log_dir = lambda: True
        return True

    def _make_env(self, base_env, codecov=True):
        env = util.Env.snapshot()

        if base_env is not None:
            base_env = dict(base_env)
            python_path = base_env.pop("PYTHONPATH", None)
            if python_path is not None:
                env.prepend_to("PYTHONPATH", python_path)
            env.update(base_env)

        env["PYTHONUNBUFFERED"] = "1"
        env["PYTHONWARNINGS"] = "error"
        env["DEBUGPY_TEST_SESSION_ID"] = str(self.id)
        env.prepend_to("PYTHONPATH", DEBUGGEE_PYTHONPATH.strpath)

        if self._init_log_dir():
            env.update(
                {
                    "DEBUGPY_LOG_DIR": self.log_dir.strpath,
                    "PYDEVD_DEBUG": "True",
                    "PYDEVD_DEBUG_FILE": (self.log_dir / "pydevd.log").strpath,
                }
            )

        if self.backchannel is not None:
            env["DEBUGPY_TEST_BACKCHANNEL_PORT"] = str(self.backchannel.port)

        if not codecov:
            # Disable codecov subprocess hook for that process.
            env.pop("COV_CORE_SOURCE", None)

        return env

    def _make_python_cmdline(self, exe, *args):
        return [
            str(s.strpath if isinstance(s, py.path.local) else s) for s in [exe, *args]
        ]

    def spawn_debuggee(self, args, cwd=None, exe=sys.executable, setup=None):
        assert self.debuggee is None
        assert not len(self.captured_output - {"stdout", "stderr"})

        args = self._make_python_cmdline(exe, *args)
        cwd = cwd.strpath if isinstance(cwd, py.path.local) else cwd

        env = self._make_env(self.spawn_debuggee.env, codecov=False)
        self.adapter_endpoints = self.tmpdir / "adapter_endpoints"
        env["DEBUGPY_ADAPTER_ENDPOINTS"] = self.adapter_endpoints.strpath
        if setup is not None:
            env["DEBUGPY_TEST_DEBUGGEE_SETUP"] = setup

        log.info(
            "Spawning {0}:\n\n"
            "Current directory: {1}\n\n"
            "Command line: {2}\n\n"
            "Environment variables: {3}\n\n",
            self.debuggee_id,
            json.repr(cwd),
            json.repr(args),
            json.repr(env),
        )

        popen_fds = {}
        capture_fds = {}
        for stream_name in self.captured_output:
            rfd, wfd = os.pipe()
            popen_fds[stream_name] = wfd
            capture_fds[stream_name] = rfd
        self.debuggee = psutil.Popen(
            args, cwd=cwd, env=env, bufsize=0, stdin=subprocess.PIPE, **popen_fds
        )
        log.info("Spawned {0} with PID={1}", self.debuggee_id, self.debuggee.pid)
        watchdog.register_spawn(self.debuggee.pid, self.debuggee_id)

        if len(capture_fds):
            self.captured_output = output.CapturedOutput(self, **capture_fds)
        for fd in popen_fds.values():
            os.close(fd)

    def wait_for_adapter_socket(self):
        log.info(
            "Waiting for {0} to open the client listener socket...", self.adapter_id
        )
        while not self.adapter_endpoints.check():
            time.sleep(0.1)

    def spawn_adapter(self, args=()):
        assert self.adapter is None
        assert self.channel is None

        args = self._make_python_cmdline(
            sys.executable, os.path.dirname(debugpy.adapter.__file__), *args
        )
        env = self._make_env(self.spawn_adapter.env)

        log.info(
            "Spawning {0}:\n\n"
            "Command line: {1}\n\n"
            "Environment variables: {2}\n\n",
            self.adapter_id,
            json.repr(args),
            json.repr(env),
        )
        self.adapter = psutil.Popen(
            args,
            bufsize=0,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            env=env,
        )
        log.info("Spawned {0} with PID={1}", self.adapter_id, self.adapter.pid)
        watchdog.register_spawn(self.adapter.pid, self.adapter_id)

        stream = messaging.JsonIOStream.from_process(self.adapter, name=self.adapter_id)
        self._start_channel(stream)

    def expect_server_socket(self, port=some.int):
        self.expected_adapter_sockets["server"] = {
            "host": some.str,
            "port": port,
            "internal": True,
        }

    def connect_to_adapter(self, address):
        assert self.channel is None

        self.before_connect(address)
        host, port = address
        log.info("Connecting to {0} at {1}:{2}", self.adapter_id, host, port)

        self.expected_adapter_sockets["client"]["port"] = port

        sock = sockets.create_client()
        sock.connect(address)

        stream = messaging.JsonIOStream.from_socket(sock, name=self.adapter_id)
        self._start_channel(stream)

    def start(self):
        config = self.config
        request = config.get("request", None)
        if request == "attach":
            host = config["connect"]["host"]
            port = config["connect"]["port"]
            self.connect_to_adapter((host, port))
            return self.request_attach()
        else:
            raise ValueError(
                f'Unsupported "request":{json.repr(request)} in session.config'
            )

    def request(self, *args, **kwargs):
        freeze = kwargs.pop("freeze", True)
        raise_if_failed = kwargs.pop("raise_if_failed", True)
        return (
            self.send_request(*args, **kwargs)
            .wait_for_response(freeze=freeze, raise_if_failed=raise_if_failed)
            .body
        )

    def send_request(self, command, arguments=None, proceed=True):
        self.before_request(command, arguments)

        if self.timeline.is_frozen and proceed:
            self.proceed()

        if command in ("launch", "attach"):
            self.start_command = command

        message = self.channel.send_request(command, arguments)
        request = self.timeline.record_request(message)

        if command in ("launch", "attach"):
            self.start_request = request

        # Register callback after recording the request, so that there's no race
        # between it being recorded, and the response to it being received.
        message.on_response(lambda response: self._process_response(request, response))

        return request

    def _process_event(self, event):
        occ = self.timeline.record_event(event, block=False)

        if event.event == "exited":
            self.observe(occ)
            self.exit_code = event("exitCode", int)
            self.exit_reason = event("reason", str, optional=True)
            assert self.exit_code == self.expected_exit_code

        elif event.event == "terminated":
            # Server socket should be closed next.
            self.expected_adapter_sockets.pop("server", None)

        elif event.event == "debugpyAttach":
            self.observe(occ)
            pid = event("subProcessId", int)
            watchdog.register_spawn(pid, f"{self.debuggee_id}-subprocess-{pid}")

        elif event.event == "debugpySockets":
            assert not self.is_subprocess
            sockets = list(event("sockets", json.array(json.object())))
            for purpose, expected_socket in self.expected_adapter_sockets.items():
                if expected_socket is None:
                    continue
                socket = None
                for socket in sockets:
                    if socket == expected_socket:
                        break
                assert (
                    socket is not None
                ), f"Expected {purpose} socket {expected_socket} not reported by adapter"
                sockets.remove(socket)
            assert not sockets, f"Unexpected sockets reported by adapter: {sockets}"

            if self.start_command == "launch":
                if "launcher" in self.expected_adapter_sockets:
                    # If adapter has just reported the launcher socket, it shouldn't be
                    # reported thereafter.
                    self.expected_adapter_sockets["launcher"] = None
                elif "server" in self.expected_adapter_sockets:
                    # If adapter just reported the server socket, the next event should
                    # report the launcher socket.
                    self.expected_adapter_sockets["launcher"] = {
                        "host": some.str,
                        "port": some.int,
                        "internal": False,
                    }

    def run_in_terminal(self, args, cwd, env):
        exe = args.pop(0)
        self.spawn_debuggee.env.update(env)
        self.spawn_debuggee(args, cwd, exe=exe)
        return {}

    def _process_request(self, request):
        self.timeline.record_request(request, block=False)
        if request.command == "runInTerminal":
            args = request("args", json.array(str, vectorize=True))
            if len(args) > 0 and request("argsCanBeInterpretedByShell", False):
                # The final arg is a string that contains multiple actual arguments.
                last_arg = args.pop()
                args += last_arg.split()
            cwd = request("cwd", ".")
            env = request("env", json.object(str))
            try:
                return self.run_in_terminal(args, cwd, env)
            except Exception as exc:
                log.swallow_exception('"runInTerminal" failed:')
                raise request.cant_handle(str(exc))

        elif request.command == "startDebugging":
            pid = request("configuration", dict)("subProcessId", int)
            watchdog.register_spawn(pid, f"{self.debuggee_id}-subprocess-{pid}")
            return {}

        else:
            raise request.isnt_valid("not supported")

    def _process_response(self, request, response):
        self.timeline.record_response(request, response, block=False)
        if request.command == "disconnect":
            # Stop the message loop, since debugpy is going to close the connection
            # from its end shortly after sending this event, and no further messages
            # are expected.
            log.info(
                'Received "disconnect" response from {0}; stopping message processing.',
                self.adapter_id,
            )
            try:
                self.channel.close()
            except Exception:
                pass

    def _process_disconnect(self):
        self.timeline.mark("disconnect", block=False)

    def _start_channel(self, stream):
        handlers = messaging.MessageHandlers(
            request=self._process_request,
            event=self._process_event,
            disconnect=self._process_disconnect,
        )
        self.channel = messaging.JsonMessageChannel(stream, handlers)
        self.channel.start()

        self.wait_for_next(
            timeline.Event(
                "output",
                {
                    "category": "telemetry",
                    "output": "ptvsd",
                    "data": {"packageVersion": some.str},
                },
            )
            & timeline.Event(
                "output",
                {
                    "category": "telemetry",
                    "output": "debugpy",
                    "data": {"packageVersion": some.str},
                },
            )
        )

        if not self.is_subprocess:
            self.wait_for_next(timeline.Event("debugpySockets"))

        self.request("initialize", self.capabilities)

    def all_events(self, event, body=some.object):
        return [
            occ.body
            for occ in self.timeline.all_occurrences_of(timeline.Event(event, body))
        ]

    def output(self, category):
        """Returns all output of a given category as a single string, assembled from
        all the "output" events received for that category so far.
        """
        events = self.all_events("output", some.dict.containing({"category": category}))
        return "".join(event("output", str) for event in events)

    def _request_start(self, method):
        self.config.normalize()
        start_request = self.send_request(method, self.config)

        # Depending on whether it's "noDebug" or not, we either get the "initialized"
        # event, or an immediate response to our request.
        self.timeline.wait_until_realized(
            timeline.Event("initialized") | timeline.Response(start_request),
            freeze=True,
        )

        if start_request.response is not None:
            # It was an immediate response - either the request failed, or there is
            # no configuration stage for this debug session.
            start_request.response.result  # raise exception if failed
            return self.wait_for_process()

        # We got "initialized" - now we need to yield to the caller, so that it can
        # configure the session before it starts running.
        return self._ConfigurationContextManager(self)

    class _ConfigurationContextManager(object):
        """Handles the start configuration sequence from "initialized" event until
        start_request receives a response.
        """

        def __init__(self, session):
            self.session = session
            self._entered = False

        def __enter__(self):
            self._entered = True
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.session.request("configurationDone")
            self.session.start_request.wait_for_response()
            self.session.wait_for_process()

        def __del__(self):
            assert self._entered, (
                "The return value of request_launch() or request_attach() must be "
                "used in a with-statement."
            )

    def request_launch(self):
        if "PYTHONPATH" in self.config.env:
            # If specified, launcher will use it in lieu of PYTHONPATH it inherited
            # from the adapter when spawning debuggee, so we need to adjust again.
            self.config.env.prepend_to("PYTHONPATH", DEBUGGEE_PYTHONPATH.strpath)

        # Adapter is going to start listening for server and spawn the launcher at
        # this point. Server socket gets reported first.
        self.expect_server_socket()

        return self._request_start("launch")

    def request_attach(self):
        # In attach(listen) scenario, adapter only starts listening for server
        # after receiving the "attach" request.
        listen = self.config.get("listen", None)
        if listen is not None:
            assert "server" not in self.expected_adapter_sockets
            self.expect_server_socket(listen["port"])
        return self._request_start("attach")

    def request_continue(self):
        self.request("continue", freeze=False)

    def set_breakpoints(self, path, lines):
        """Sets breakpoints in the specified file, and returns the list of all the
        corresponding DAP Breakpoint objects in the same order.

        If lines are specified, it should be an iterable in which every element is
        either a line number or a string. If it is a string, then it is translated
        to the corresponding line number via get_marked_line_numbers(path).

        If lines=all, breakpoints will be set on all the marked lines in the file.
        """

        # Don't fetch line markers unless needed - in some cases, the breakpoints
        # might be set in a file that does not exist on disk (e.g. remote attach).
        get_marked_line_numbers = lambda: code.get_marked_line_numbers(path)

        if lines is all:
            lines = get_marked_line_numbers().keys()

        def make_breakpoint(line):
            if isinstance(line, int):
                descr = str(line)
            else:
                marker = line
                line = get_marked_line_numbers()[marker]
                descr = f"{line} (@{marker})"
            bp_log.append((line, descr))
            return {"line": line}

        bp_log = []
        breakpoints = self.request(
            "setBreakpoints",
            {
                "source": {"path": path},
                "breakpoints": [make_breakpoint(line) for line in lines],
            },
        )("breakpoints", json.array())

        bp_log = sorted(bp_log, key=lambda pair: pair[0])
        bp_log = ", ".join((descr for _, descr in bp_log))
        log.info("Breakpoints set in {0}: {1}", path, bp_log)

        return breakpoints

    def get_variables(self, *varnames, **kwargs):
        """Fetches the specified variables from the frame specified by frame_id, or
        from the topmost frame in the last "stackTrace" response if frame_id is not
        specified.

        If varnames is empty, then all variables in the frame are returned. The result
        is an OrderedDict, in which every entry has variable name as the key, and a
        DAP Variable object as the value. The original order of variables as reported
        by the debugger is preserved.

        If varnames is not empty, then only the specified variables are returned.
        The result is a tuple, in which every entry is a DAP Variable object; those
        entries are in the same order as varnames.
        """

        assert self.timeline.is_frozen

        frame_id = kwargs.pop("frame_id", None)
        if frame_id is None:
            stackTrace_responses = self.all_occurrences_of(
                timeline.Response(timeline.Request("stackTrace"))
            )
            assert stackTrace_responses, (
                "get_variables() without frame_id requires at least one response "
                'to a "stackTrace" request in the timeline.'
            )
            stack_trace = stackTrace_responses[-1]
            frame_id = stack_trace.body.get("stackFrames", json.array())[0]("id", int)

        scopes = self.request("scopes", {"frameId": frame_id})("scopes", json.array())
        assert len(scopes) > 0

        variables = self.request(
            "variables", {"variablesReference": scopes[0]("variablesReference", int)}
        )("variables", json.array())

        variables = collections.OrderedDict(((v("name", str), v) for v in variables))
        if varnames:
            assert set(varnames) <= set(variables.keys())
            return tuple((variables[name] for name in varnames))
        else:
            return variables

    def get_variable(self, varname, frame_id=None):
        """Same as get_variables(...)[0]."""
        return self.get_variables(varname, frame_id=frame_id)[0]

    def wait_for_next_event(self, event, body=some.object, freeze=True):
        return self.timeline.wait_for_next(
            timeline.Event(event, body), freeze=freeze
        ).body

    def wait_for_process(self):
        process = self.wait_for_next_event("process", freeze=False)
        assert process == some.dict.containing(
            {
                "startMethod": self.start_request.command,
                "name": some.str,
                "isLocalProcess": True,
                "systemProcessId": some.int,
            }
        )

    def wait_for_stop(
        self,
        reason=some.str,
        expected_frames=None,
        expected_text=None,
        expected_description=None,
    ):
        stopped = self.wait_for_next_event("stopped")

        expected_stopped = {
            "reason": reason,
            "threadId": some.int,
            "allThreadsStopped": True,
        }
        if expected_text is not None:
            expected_stopped["text"] = expected_text
        if expected_description is not None:
            expected_stopped["description"] = expected_description
        if stopped("reason", str) not in [
            "step",
            "exception",
            "breakpoint",
            "entry",
            "goto",
        ]:
            expected_stopped["preserveFocusHint"] = True
        assert stopped == some.dict.containing(expected_stopped)

        tid = stopped("threadId", int)
        stack_trace = self.request("stackTrace", {"threadId": tid})
        frames = stack_trace("stackFrames", json.array()) or []
        assert len(frames) == stack_trace("totalFrames", int)

        if expected_frames:
            assert len(expected_frames) <= len(frames)
            assert expected_frames == frames[0 : len(expected_frames)]

        fid = frames[0]("id", int)
        return StopInfo(stopped, frames, tid, fid)

    def wait_for_next_subprocess(self):
        message = self.timeline.wait_for_next(
            timeline.Event("debugpyAttach") | timeline.Request("startDebugging")
        )
        if isinstance(message, timeline.EventOccurrence):
            config = message.body
            assert "request" in config
        elif isinstance(message, timeline.RequestOccurrence):
            config = dict(message.body("configuration", dict))
            assert "request" not in config
            config["request"] = "attach"
        return Session(config)

    def wait_for_disconnect(self):
        self.timeline.wait_until_realized(timeline.Mark("disconnect"), freeze=True)

    def wait_for_exit(self):
        if self.debuggee is not None:
            log.info("Waiting for {0} to exit ...", self.debuggee_id)
            try:
                self.debuggee.wait()
            except Exception:
                pass
            finally:
                watchdog.unregister_spawn(self.debuggee.pid, self.debuggee_id)

        self.wait_for_terminated()

        # FIXME: "exited" event is not properly reported in attach scenarios at the
        # moment, so the exit code is only checked if it's present.
        if self.debuggee is not None and self.exit_code is not None:
            assert self.debuggee.returncode == self.exit_code
        return self.exit_code

    def wait_for_terminated(self):
        self.timeline.wait_until_realized(timeline.Event("terminated"))

    def captured_stdout(self, encoding=None):
        assert self.debuggee is not None
        return self.captured_output.stdout(encoding)

    def captured_stderr(self, encoding=None):
        assert self.debuggee is not None
        return self.captured_output.stderr(encoding)

    def disconnect(self, force=False):
        if self.channel is None:
            return

        try:
            if not force:
                self.request("disconnect")
                self.wait_for_terminated()
        except messaging.JsonIOError:
            pass
        finally:
            try:
                self.channel.close()
            except Exception:
                pass
            self.channel.wait()
            self.channel = None


Session.reset_counter()



# coding: utf-8
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import pytest
import re
import sys

import debugpy
from debugpy.common import messaging
from tests import debug, test_data, timeline
from tests.debug import runners, targets
from tests.patterns import some


@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("target", targets.all)
@pytest.mark.flaky(retries=2, delay=1)
def test_run(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import os
        import sys

        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        print("begin")
        backchannel.send(os.path.abspath(sys.modules["debugpy"].__file__))
        assert backchannel.receive() == "continue"
        print("end")

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        expected_debugpy_path = os.path.abspath(debugpy.__file__)
        assert backchannel.receive() == some.str.matching(
            re.escape(expected_debugpy_path) + r"(c|o)?"
        )

        backchannel.send("continue")
        session.wait_for_next_event("terminated")
        session.proceed()


@pytest.mark.parametrize("run", [runners.launch["internalConsole"]])
def test_run_relative_path(pyfile, run):
    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel
        from _pydev_bundle.pydev_log import list_log_files

        debuggee.setup()
        from _pydevd_bundle import pydevd_constants  # @ bp1

        backchannel.send(
            list_log_files(pydevd_constants.DebugInfoHolder.PYDEVD_DEBUG_FILE)
        )
        assert backchannel.receive() == "continue"

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        code_to_debug = str(code_to_debug)
        cwd = os.path.dirname(os.path.dirname(code_to_debug))

        program = targets.Program(code_to_debug)
        program.make_relative(cwd)
        with run(session, program):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop()
        session.request_continue()

        pydevd_debug_files = backchannel.receive()
        backchannel.send("continue")
        session.wait_for_next_event("terminated")
        session.proceed()

    # Check if we don't have errors in the pydevd log (the
    # particular error this test is covering:
    # https://github.com/microsoft/debugpy/issues/620
    # is handled by pydevd but produces a Traceback in the logs).
    for pydevd_debug_file in pydevd_debug_files:
        with open(pydevd_debug_file, "r") as stream:
            contents = stream.read()

    assert "FileNotFound" not in contents


@pytest.mark.parametrize("run", runners.all_launch)
def test_run_submodule(run):
    with debug.Session() as session:
        session.config["cwd"] = test_data / "testpkgs"

        backchannel = session.open_backchannel()
        with run(session, targets.Module(name="pkg1.sub")):
            pass

        assert backchannel.receive() == "ok"


@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.parametrize("target", targets.all)
def test_nodebug(pyfile, run, target):
    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.receive()  # @ bp1
        print("ok")  # @ bp2

    with debug.Session() as session:
        session.config["noDebug"] = True
        session.config["redirectOutput"] = True

        backchannel = session.open_backchannel()
        run(session, target(code_to_debug))

        with pytest.raises(messaging.MessageHandlingError):
            session.set_breakpoints(code_to_debug, all)

        backchannel.send(None)

        # Breakpoint shouldn't be hit.
        pass

    assert "ok" in session.output("stdout")


@pytest.mark.skipif(sys.platform == "win32", reason="sudo not available on Windows")
@pytest.mark.parametrize("run", runners.all_launch)
def test_sudo(pyfile, tmpdir, run, target):
    # Since the test can't rely on sudo being allowed for the user, create a dummy
    # sudo script that doesn't actually elevate, but sets an environment variable
    # that can be checked in the debuggee.
    sudo = tmpdir / "sudo"
    sudo.write(
        """#!/bin/sh
        if [ "$1" = "-E" ]; then shift; fi
        exec env DEBUGPY_SUDO=1 "$@"
        """
    )
    os.chmod(sudo.strpath, 0o777)

    @pyfile
    def code_to_debug():
        import os

        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(os.getenv("DEBUGPY_SUDO", "0"))

    with debug.Session() as session:
        session.config["sudo"] = True
        session.spawn_adapter.env["PATH"] = session.spawn_debuggee.env["PATH"] = (
            tmpdir.strpath + ":" + os.environ["PATH"]
        )

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        # The "runInTerminal" request sent by the adapter to spawn the launcher,
        # if any, shouldn't be using sudo.
        assert all(
            "sudo" not in req.arguments["args"]
            for req in session.all_occurrences_of(timeline.Request("runInTerminal"))
        )

        # The launcher, however, should use our dummy sudo to spawn the debuggee,
        # and the debuggee should report the environment variable accordingly.
        assert backchannel.receive() == "1"


def make_custompy(tmpdir, id=""):
    if sys.platform == "win32":
        custompy = tmpdir / ("custompy" + id + ".cmd")
        script = """@echo off
            set DEBUGPY_CUSTOM_PYTHON=<id>;%DEBUGPY_CUSTOM_PYTHON%
            <python> %*
            """
    else:
        custompy = tmpdir / ("custompy" + id + ".sh")
        script = """#!/bin/sh
            env "DEBUGPY_CUSTOM_PYTHON=<id>;$DEBUGPY_CUSTOM_PYTHON" <python> "$@"
            """

    script = script.replace("<id>", id)
    script = script.replace("<python>", sys.executable)
    custompy.write(script)
    os.chmod(custompy.strpath, 0o777)

    return custompy.strpath


@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.parametrize("debuggee_custompy", [None, "launcher"])
@pytest.mark.parametrize("launcher_custompy", [None, "debuggee"])
def test_custom_python(
    pyfile, tmpdir, run, target, debuggee_custompy, launcher_custompy
):
    @pyfile
    def code_to_debug():
        import os

        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(os.getenv("DEBUGPY_CUSTOM_PYTHON"))

    expected = ""
    if debuggee_custompy:
        debuggee_custompy = make_custompy(tmpdir, "debuggee")
        expected += "debuggee;"
    if launcher_custompy:
        launcher_custompy = make_custompy(tmpdir, "launcher")
        expected += "launcher;"
    else:
        # If "python" is set, it also becomes the default for "debugLauncherPython"
        expected *= 2
    if not len(expected):
        pytest.skip()

    with debug.Session() as session:
        session.config.pop("python", None)
        if launcher_custompy:
            session.config["debugLauncherPython"] = launcher_custompy
        if debuggee_custompy:
            session.config["python"] = debuggee_custompy

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        assert backchannel.receive() == expected


@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.parametrize("python_args", [None, "-B"])
@pytest.mark.parametrize("python", [None, "custompy", "custompy,-O"])
@pytest.mark.parametrize("python_key", ["python", "pythonPath"])
@pytest.mark.flaky(retries=2, delay=1)
def test_custom_python_args(
    pyfile, tmpdir, run, target, python_key, python, python_args
):
    @pyfile
    def code_to_debug():
        import sys

        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send([sys.flags.optimize, sys.flags.dont_write_bytecode])

    custompy = make_custompy(tmpdir)
    python = [] if python is None else python.split(",")
    python = [(custompy if arg == "custompy" else arg) for arg in python]
    python_args = [] if python_args is None else python_args.split(",")
    python_cmd = (python if len(python) else [sys.executable]) + python_args

    with debug.Session() as session:
        session.config.pop("python", None)
        session.config.pop("pythonPath", None)
        if len(python):
            session.config[python_key] = python[0] if len(python) == 1 else python
        if len(python_args):
            session.config["pythonArgs"] = python_args

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        assert backchannel.receive() == ["-O" in python_cmd, "-B" in python_cmd]


@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("target", targets.all)
@pytest.mark.parametrize("frame_eval", ["", "yes", "no"])
def test_frame_eval(pyfile, target, run, frame_eval):
    # Frame-eval optimizations are not available for some Python implementations,
    # but pydevd will still try to use them if the environment variable is set to
    # "yes" explicitly, so the test must detect and skip those cases.
    if frame_eval == "yes":
        try:
            import _pydevd_frame_eval.pydevd_frame_eval_cython_wrapper  # noqa
        except ImportError:
            pytest.skip("Frame-eval not available")
        else:
            pass

    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel

        debuggee.setup()

        from _pydevd_frame_eval.pydevd_frame_eval_main import USING_FRAME_EVAL

        backchannel.send(USING_FRAME_EVAL)

    with debug.Session() as session:
        assert "PYDEVD_USE_FRAME_EVAL" not in os.environ
        if len(frame_eval):
            env = (
                session.config.env
                if run.request == "launch"
                else session.spawn_debuggee.env
            )
            env["PYDEVD_USE_FRAME_EVAL"] = frame_eval

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        using_frame_eval = backchannel.receive()
        assert using_frame_eval == (frame_eval == "yes")


@pytest.mark.parametrize("run", [runners.all_launch[0]])
def test_unicode_dir(tmpdir, run, target):
    unicode_chars = "á"

    directory = os.path.join(str(tmpdir), unicode_chars)
    os.makedirs(directory)

    code_to_debug = os.path.join(directory, "experiment.py")
    with open(code_to_debug, "wb") as stream:
        stream.write(
            b"""
import debuggee
from debuggee import backchannel

debuggee.setup()
backchannel.send('ok')
"""
        )

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            pass

        received = backchannel.receive()
        assert received == "ok"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

from tests import code, debug, log, net, test_data
from tests.debug import runners, targets
from tests.patterns import some

pytestmark = pytest.mark.timeout(60)

flask_server = net.WebServer(net.get_test_server_port())


class paths:
    flask1 = test_data / "flask1"
    app_py = flask1 / "app.py"
    main_py = flask1 / "main.py"
    hello_html = flask1 / "templates" / "hello.html"
    bad_html = flask1 / "templates" / "bad.html"


class lines:
    app_py = code.get_marked_line_numbers(paths.app_py)


@pytest.fixture
@pytest.mark.parametrize("run", [runners.launch, runners.attach_connect["cli"]])
def start_flask(run):
    def start(session, multiprocess=False):
        # No clean way to kill Flask server, expect non-zero exit code
        session.expected_exit_code = some.int

        session.config.env.update(
            {
                "FLASK_APP": paths.app_py,
                "FLASK_ENV": "development",
                "FLASK_DEBUG": "1" if multiprocess else "0",
            }
        )
        if sys.platform != "win32":
            locale = "en_US.utf8" if sys.platform.startswith("linux") else "en_US.UTF-8"
            session.config.env.update({"LC_ALL": locale, "LANG": locale})

        session.config.update({"jinja": True, "subProcess": multiprocess})

        args = ["run", "--no-debugger", "--with-threads"]
        if multiprocess:
            args += ["--reload"]
        else:
            args += ["--no-reload"]
        args += ["--port", str(flask_server.port)]

        if multiprocess and run.request == "attach":
            # For multiproc attach, we need to use a helper stub to import debuggee
            # before running Flask; otherwise, we will get the connection only from
            # the subprocess, not from the Flask server process.
            target = targets.Program(paths.main_py, args=args)
        else:
            target = targets.Module(name="flask", args=args)

        return run(session, target, cwd=paths.flask1)

    return start


@pytest.mark.parametrize("bp_target", ["code", "template"])
def test_flask_breakpoint_no_multiproc(start_flask, bp_target):
    bp_file, bp_line, bp_name = {
        "code": (paths.app_py, lines.app_py["bphome"], "home"),
        "template": (paths.hello_html, 8, "template"),
    }[bp_target]
    bp_var_content = "Flask-Jinja-Test"

    with debug.Session() as session:
        with start_flask(session):
            breakpoints = session.set_breakpoints(bp_file, [bp_line])
            for bp in breakpoints:
                # They'll be verified later on for templates.
                assert bp["verified"] == (bp_target == "code")

        with flask_server:
            home_request = flask_server.get("/")

            if bp_target == "template":
                breakpoint_body = session.wait_for_next_event("breakpoint")
                assert breakpoint_body["reason"] == "changed"
                assert breakpoint_body["breakpoint"]["verified"]

            session.wait_for_stop(
                "breakpoint",
                expected_frames=[
                    some.dap.frame(some.dap.source(bp_file), name=bp_name, line=bp_line)
                ],
            )

            var_content = session.get_variable("content")
            assert var_content == some.dict.containing(
                {
                    "name": "content",
                    "type": "str",
                    "value": repr(bp_var_content),
                    "presentationHint": {"attributes": ["rawString"]},
                    "evaluateName": "content",
                    "variablesReference": 0,
                }
            )

            session.request_continue()
            assert bp_var_content in home_request.response_text()


def test_flask_template_exception_no_multiproc(start_flask):
    with debug.Session() as session:
        with start_flask(session):
            session.request(
                "setExceptionBreakpoints", {"filters": ["raised", "uncaught"]}
            )

        with flask_server:
            flask_server.get("/badtemplate")
            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(paths.bad_html),
                        name=some.str,  # varies depending on Jinja version
                        line=8,
                    )
                ],
            )

            exception_info = session.request(
                "exceptionInfo", {"threadId": stop.thread_id}
            )
            assert exception_info == some.dict.containing(
                {
                    "exceptionId": some.str.ending_with("TemplateSyntaxError"),
                    "breakMode": "always",
                    "description": some.str.containing("doesnotexist"),
                    "details": some.dict.containing(
                        {
                            "message": some.str.containing("doesnotexist"),
                            "typeName": some.str.ending_with("TemplateSyntaxError"),
                        }
                    ),
                }
            )

            # Exception gets reported again as it is re-raised in every frame between
            # the template and app.py. The number of frames is a Flask implementation
            # detail, and varies between versions, so we keep iterating until we see
            # it reported in app.py.
            while True:
                log.info("Exception propagating to next frame...")
                session.request_continue()
                stop = session.wait_for_stop("exception")
                if stop.frames[0] == some.dap.frame(paths.app_py, line=some.int):
                    break

            # Let the request finish processing and respond with HTTP 500.
            session.request_continue()


@pytest.mark.parametrize("exc_type", ["handled", "unhandled"])
def test_flask_exception_no_multiproc(start_flask, exc_type):
    exc_line = lines.app_py["exc_" + exc_type]

    with debug.Session() as session:
        with start_flask(session):
            session.request(
                "setExceptionBreakpoints", {"filters": ["raised", "uncaught"]}
            )

        with flask_server:
            flask_server.get("/" + exc_type)
            stopped = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(paths.app_py),
                        line=exc_line,
                        name="bad_route_" + exc_type,
                    )
                ],
            ).body

            assert stopped == some.dict.containing(
                {
                    "reason": "exception",
                    "text": some.str.ending_with("ArithmeticError"),
                    "description": "Hello",
                }
            )

            exception_info = session.request(
                "exceptionInfo", {"threadId": stopped["threadId"]}
            )

            assert exception_info == {
                "exceptionId": some.str.ending_with("ArithmeticError"),
                "breakMode": "always",
                "description": "Hello",
                "details": {
                    "message": "Hello",
                    "typeName": some.str.ending_with("ArithmeticError"),
                    "source": some.path(paths.app_py),
                    "stackTrace": some.str,
                },
            }

            session.request_continue()


def test_flask_breakpoint_multiproc(start_flask):
    bp_line = lines.app_py["bphome"]
    bp_var_content = "Flask-Jinja-Test"

    with debug.Session() as parent_session:
        with start_flask(parent_session, multiprocess=True):
            parent_session.set_breakpoints(paths.app_py, [bp_line])

        with parent_session.wait_for_next_subprocess() as child_session:
            with child_session.start():
                child_session.set_breakpoints(paths.app_py, [bp_line])

            with flask_server:
                home_request = flask_server.get("/")
                child_session.wait_for_stop(
                    "breakpoint",
                    expected_frames=[
                        some.dap.frame(
                            some.dap.source(paths.app_py), line=bp_line, name="home"
                        )
                    ],
                )

                var_content = child_session.get_variable("content")
                assert var_content == some.dict.containing(
                    {
                        "name": "content",
                        "type": "str",
                        "value": repr(bp_var_content),
                        "presentationHint": {"attributes": ["rawString"]},
                        "evaluateName": "content",
                        "variablesReference": 0,
                    }
                )

                child_session.request_continue()
                assert bp_var_content in home_request.response_text()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Makes sure that the code is run under debugger, using the appropriate method
to establish connection back to DebugSession in the test process, depending on
DebugSession.start_method used by the test.

This module MUST be imported, and setup() must be called, by all test scripts that
are run via debug.Session and the standard runners.
"""

__all__ = ["debugpy", "pydevd", "session_id"]

import os


# Used by backchannel.
session_id = int(os.getenv("DEBUGPY_TEST_SESSION_ID"))
name = "Debuggee-" + str(session_id)


# For non-blocking communication between the test and the debuggee. The debuggee
# can access this as a normal dict - scratchpad["foo"] etc. The test should assign
# to session.scratchpad[...], which will automatically perform "evaluate" requests
# as needed to assign the value.
scratchpad = {}


# Some runners require code to be executed in the debuggee process, either to set up
# the debug server, or to ensure that it doesn't run any other code until the debugger
# is attached. This provides a facility to inject such code.
def setup():
    _code = os.environ.pop("DEBUGPY_TEST_DEBUGGEE_SETUP", None)
    if _code:
        _code = compile(_code, "<DEBUGPY_TEST_DEBUGGEE_SETUP>", "exec")
        eval(_code, {})



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import contextlib
import pytest

from tests import debug
from tests.debug import runners, targets


@contextlib.contextmanager
def check_logs(tmpdir, run):
    # For attach_pid, there's ptvsd.server process that performs the injection,
    # and then there's the debug server that is injected into the debuggee.
    server_count = 2 if type(run).__name__ == "attach_pid" else 1

    expected_logs = {
        "debugpy.adapter-*.log": 1,
        "debugpy.launcher-*.log": 1 if run.request == "launch" else 0,
        "debugpy.pydevd.*.log": server_count,
        "debugpy.server-*.log": server_count,
    }

    actual_logs = lambda: {
        filename: len(tmpdir.listdir(filename)) for filename in expected_logs
    }

    assert actual_logs() == {filename: 0 for filename in expected_logs}
    yield
    assert actual_logs() == expected_logs


@pytest.mark.parametrize("run", runners.all_attach_socket)
@pytest.mark.parametrize("target", targets.all)
def test_log_dir(pyfile, tmpdir, run, target):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

    # Depending on the method, the runner will use either `debugpy --log-dir ...`
    # or `debugpy.log_to() ...`.
    run = run.with_options(log_dir=tmpdir.strpath)
    with check_logs(tmpdir, run):
        with debug.Session() as session:
            session.log_dir = None

            with run(session, target(code_to_debug)):
                pass


@pytest.mark.parametrize("run", runners.all)
@pytest.mark.parametrize("target", targets.all)
def test_log_dir_env(pyfile, tmpdir, run, target):
    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        assert backchannel.receive() == "proceed"

    with check_logs(tmpdir, run):
        with debug.Session() as session:
            session.log_dir = None
            session.spawn_adapter.env["DEBUGPY_LOG_DIR"] = tmpdir.strpath
            if run.request != "launch":
                session.spawn_debuggee.env["DEBUGPY_LOG_DIR"] = tmpdir.strpath

            backchannel = session.open_backchannel()
            with run(session, target(code_to_debug)):
                pass

            backchannel.send("proceed")



import unittest


class SetUpClassTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        raise ValueError("This is an INTENTIONAL value error in setUpClass.")

    def test_blank(self):
        pass


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import pytest

from tests import debug
from tests.debug import runners
from tests.patterns import some


@pytest.mark.parametrize("run", runners.all_attach_socket)
def test_continue_on_disconnect_for_attach(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send("continued")  # @bp

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(code_to_debug, line="bp")]
        )
        session.disconnect()
        assert "continued" == backchannel.receive()


@pytest.mark.parametrize("run", runners.all_launch)
def test_exit_on_disconnect_for_launch(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys

        debuggee.setup()
        filename = sys.argv[1]  # @bp
        # Disconnect happens here; subsequent lines should not run.
        with open(filename, "w") as f:
            f.write("failed")

    filename = (code_to_debug.dirpath() / "failed.txt").strpath

    with debug.Session() as session:
        session.expected_exit_code = some.int
        with run(session, target(code_to_debug, args=[filename])):
            session.set_breakpoints(code_to_debug, all)

        session.wait_for_stop(
            "breakpoint", expected_frames=[some.dap.frame(code_to_debug, line="bp")]
        )
        session.disconnect()

    assert not os.path.exists(filename)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from debugpy.common import json

import tests
from tests import code, debug, log, test_data
from tests.debug import targets
from tests.patterns import some


if not tests.full:

    @pytest.fixture(params=targets.all_named)
    def target(request):
        return request.param


@pytest.mark.parametrize("scenario", ["exclude_by_name", "exclude_by_dir"])
@pytest.mark.parametrize("exc_type", ["RuntimeError", "SystemExit"])
def test_exceptions_and_exclude_rules(pyfile, target, run, scenario, exc_type):
    if exc_type == "RuntimeError":

        @pyfile
        def code_to_debug():
            import debuggee

            debuggee.setup()
            raise RuntimeError("unhandled error")  # @raise_line

    elif exc_type == "SystemExit":

        @pyfile
        def code_to_debug():
            import debuggee
            import sys

            debuggee.setup()
            sys.exit(1)  # @raise_line

    else:
        pytest.fail(exc_type)

    if scenario == "exclude_by_name":
        rules = [{"path": "**/" + code_to_debug.basename, "include": False}]
    elif scenario == "exclude_by_dir":
        rules = [{"path": code_to_debug.dirname, "include": False}]
    else:
        pytest.fail(scenario)
    log.info("Rules: {0}", json.repr(rules))

    with debug.Session() as session:
        session.expected_exit_code = some.int
        session.config["rules"] = rules

        with run(session, target(code_to_debug)):
            session.request(
                "setExceptionBreakpoints", {"filters": ["raised", "uncaught"]}
            )

        # No exceptions should be seen.
        session.wait_for_next_event("terminated")
        session.proceed()


@pytest.mark.parametrize("scenario", ["exclude_code_to_debug", "exclude_callback_dir"])
def test_exceptions_and_partial_exclude_rules(pyfile, target, run, scenario):
    @pyfile
    def code_to_debug():
        import debuggee
        import sys
        from debuggee import backchannel

        debuggee.setup()
        call_me_back_dir = backchannel.receive()
        sys.path.insert(0, call_me_back_dir)

        import call_me_back

        def call_func():
            raise RuntimeError("unhandled error")  # @raise

        call_me_back.call_me_back(call_func)  # @call_me_back
        print("done")

    call_me_back_dir = test_data / "call_me_back"
    call_me_back_py = call_me_back_dir / "call_me_back.py"
    call_me_back_py.lines = code.get_marked_line_numbers(call_me_back_py)

    if scenario == "exclude_code_to_debug":
        rules = [{"path": "**/" + code_to_debug.basename, "include": False}]
    elif scenario == "exclude_callback_dir":
        rules = [{"path": call_me_back_dir, "include": False}]
    else:
        pytest.fail(scenario)
    log.info("Rules: {0}", json.repr(rules))

    with debug.Session() as session:
        session.expected_exit_code = some.int
        session.config["rules"] = rules

        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug)):
            session.request(
                "setExceptionBreakpoints", {"filters": ["raised", "uncaught"]}
            )

        backchannel.send(call_me_back_dir)

        if scenario == "exclude_code_to_debug":
            # Stop at handled exception, with code_to_debug.py excluded.
            #
            # Since the module raising the exception is excluded, it must not stop at
            # @raise, but rather at @callback (i.e. the closest non-excluded frame).

            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(call_me_back_py),
                        line=call_me_back_py.lines["callback"],
                    )
                ],
            )
            assert stop.frames != some.list.containing(
                [some.dap.frame(some.dap.source(code_to_debug), line=some.int)]
            )

            # As exception unwinds the stack, we shouldn't stop at @call_me_back,
            # since that line is in the excluded file.
            #
            # Afterwards, because the exception unhandled, we'll have an additional stop
            # (although the unhandled exception is last seen in an excluded file, we'll
            # show it if it has a non-excluded file in the stack).
            session.request_continue()

            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(call_me_back_py),
                        line=call_me_back_py.lines["callback"],
                    )
                ],
            )
            session.request_continue()

        elif scenario == "exclude_callback_dir":
            # Stop at handled exception, with call_me_back.py excluded.
            #
            # Since the module raising the exception is not excluded, it must stop at
            # @raise.

            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(code_to_debug),
                        name="call_func",
                        line=code_to_debug.lines["raise"],
                    ),
                    some.dap.frame(
                        some.dap.source(code_to_debug),
                        name="<module>",
                        line=code_to_debug.lines["call_me_back"],
                    ),
                ],
            )
            assert stop.frames != some.list.containing(
                [some.dap.frame(some.dap.source(call_me_back_py), line=some.int)]
            )

            session.request_continue()

            # As exception unwinds the stack, it must not stop at @callback, since that
            # line is in the excluded file. However, it must stop at @call_me_back.
            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(code_to_debug),
                        name="<module>",
                        line=code_to_debug.lines["call_me_back"],
                    )
                ],
            )
            assert stop.frames != some.list.containing(
                [some.dap.frame(some.dap.source(call_me_back_py), line=some.int)]
            )

            session.request_continue()

            # Now the exception is unhandled, and should be reported as such.
            stop = session.wait_for_stop(
                "exception",
                expected_frames=[
                    some.dap.frame(
                        some.dap.source(code_to_debug),
                        name="call_func",
                        line=code_to_debug.lines["raise"],
                    ),
                    some.dap.frame(
                        some.dap.source(code_to_debug),
                        name="<module>",
                        line=code_to_debug.lines["call_me_back"],
                    ),
                ],
            )
            assert stop.frames != some.list.containing(
                [some.dap.frame(some.dap.source(call_me_back_py), line=some.int)]
            )

            # Let the process crash due to unhandled exception.
            session.request_continue()

        else:
            pytest.fail(scenario)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import pytest
import pytest_timeout
import sys

from debugpy.common import log # pyright: ignore[reportAttributeAccessIssue]
import tests
from tests import logs


def pytest_addoption(parser):
    parser.addoption(
        "--debugpy-log-dir",
        type=str,
        help="Write debugpy and pydevd logs to the specified directory",
    )
    parser.addoption(
        "--debugpy-log-passed",
        action="store_true",
        help="Keep debugpy and pydevd logs for tests that passed",
    )


def pytest_configure(config):
    if config.option.debugpy_log_dir:
        log.log_dir = config.option.debugpy_log_dir
    else:
        bits = 64 if sys.maxsize > 2**32 else 32
        ver = "{0}.{1}-{bits}".format(*sys.version_info, bits=bits)
        log.log_dir = (tests.root / "_logs" / ver).strpath
    log.info("debugpy and pydevd logs will be under {0}", log.log_dir)


def pytest_report_header(config):
    return log.get_environment_description(f"Test environment for tests-{os.getpid()}")


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_runtest_makereport(item, call):
    # Adds attributes setup_report, call_report, and teardown_report to the item,
    # referencing TestReport instances for the corresponding phases, after the scope
    # finished running its tests. This can be used in function-level fixtures to
    # detect test failures, e.g.:
    #
    #   if request.node.call_report.failed: ...

    outcome = yield
    report = outcome.get_result()
    setattr(item, report.when + "_report", report)


def pytest_make_parametrize_id(config, val):
    return getattr(val, "pytest_id", None)

# If a test times out and pytest tries to print the stacks of where it was hanging,
# we want to print the pydevd log as well. This is not a normal pytest hook - we
# just detour pytest_timeout.dump_stacks directly.
_dump_stacks = pytest_timeout.dump_stacks
pytest_timeout.dump_stacks = lambda terminal: (_dump_stacks(terminal), logs.dump())



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from tests import debug
from tests.patterns import some


def test_set_next_statement(pyfile, run, target):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        def func():
            print(1)  # @inner1
            print(2)  # @inner2

        print(3)  # @outer3
        func()

    line_numbers = code_to_debug.lines

    with debug.Session() as session:
        with run(session, target(code_to_debug)):
            session.set_breakpoints(code_to_debug, ["inner1"])

        stop = session.wait_for_stop(
            expected_frames=[some.dap.frame(code_to_debug, "inner1")]
        )

        targets = session.request(
            "gotoTargets",
            {"source": {"path": code_to_debug}, "line": line_numbers["outer3"]},
        )["targets"]
        assert targets == [
            {"id": some.number, "label": some.str, "line": line_numbers["outer3"]}
        ]
        outer3_target = targets[0]["id"]

        with pytest.raises(Exception):
            session.request(
                "goto", {"threadId": stop.thread_id, "targetId": outer3_target}
            )

        targets = session.request(
            "gotoTargets",
            {"source": {"path": code_to_debug}, "line": line_numbers["inner2"]},
        )["targets"]
        assert targets == [
            {"id": some.number, "label": some.str, "line": line_numbers["inner2"]}
        ]
        inner2_target = targets[0]["id"]

        session.request("goto", {"threadId": stop.thread_id, "targetId": inner2_target})
        session.wait_for_next_event("continued")

        stop = session.wait_for_stop(
            "goto", expected_frames=[some.dap.frame(code_to_debug, "inner2")]
        )
        session.request_continue()



import debuggee

debuggee.setup()

import os
import sys

from django.conf import settings
from django.core.management import execute_from_command_line
from django.core.signals import request_finished
from django.dispatch import receiver
from django.http import HttpResponse
from django.template import loader


exiting = False


@receiver(request_finished)
def on_request_finished(sender, **kwargs):
    if exiting:
        os._exit(0)


settings.configure(
    MIDDLEWARE=[],
    DEBUG=True,
    SECRET_KEY="Placeholder_CD8FF4C1-7E6C-4E45-922D-C796271F2345",
    ROOT_URLCONF=sys.modules[__name__],
    SETTINGS_MODULE="",  # Added to avoid a KeyError during shutdown on the bad template test.
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "APP_DIRS": True,
            "DIRS": [os.path.join(os.path.dirname(__file__), "templates")],
        }
    ],
)


def home(request):
    title = "hello"
    content = "Django-Django-Test"
    template = loader.get_template("hello.html")  # @bphome
    context = {"title": title, "content": content}
    return HttpResponse(template.render(context, request))


def bad_route_handled(request):
    try:
        raise ArithmeticError("Hello")  # @exc_handled
    except Exception:
        pass
    title = "hello"
    content = "Django-Django-Test"
    template = loader.get_template("hello.html")
    context = {"title": title, "content": content}
    return HttpResponse(template.render(context, request))


def bad_route_unhandled(request):
    raise ArithmeticError("Hello")  # @exc_unhandled
    title = "hello"
    content = "Django-Django-Test"
    template = loader.get_template("hello.html")
    context = {"title": title, "content": content}
    return HttpResponse(template.render(context, request))


def bad_template(request):
    title = "hello"
    content = "Django-Django-Test"
    context = {"title": title, "content": content}
    template = loader.get_template("bad.html")
    return HttpResponse(template.render(context, request))


def exit_app(request):
    global exiting
    exiting = True
    return HttpResponse("Done")


from django.urls import path

urlpatterns = [
    path("home", home, name="home"),
    path("handled", bad_route_handled, name="bad_route_handled"),
    path("unhandled", bad_route_unhandled, name="bad_route_unhandled"),
    path("badtemplate", bad_template, name="bad_template"),
    path("exit", exit_app, name="exit_app"),
]

if __name__ == "__main__":
    execute_from_command_line(sys.argv)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest

from debugpy.common import log
from tests import debug
from tests.debug import runners, targets
from tests.patterns import some


@pytest.mark.parametrize("target", targets.all)
@pytest.mark.parametrize("run", runners.all)
def test_args(pyfile, target, run):
    @pyfile
    def code_to_debug():
        import sys
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(sys.argv)

    args = ["--arg1", "arg2", "-arg3", "--", "arg4", "-a"]

    with debug.Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug, args=args)):
            pass
        argv = backchannel.receive()
        assert argv == [some.str] + args


@pytest.mark.parametrize("target", targets.all)
@pytest.mark.parametrize("run", runners.all_launch)
@pytest.mark.parametrize("expansion", ["preserve", "expand"])
def test_shell_expansion(pyfile, target, run, expansion):
    if expansion == "expand" and run.console == "internalConsole":
        pytest.skip('Shell expansion is not supported for "internalConsole"')

    @pyfile
    def code_to_debug():
        import sys
        import debuggee
        from debuggee import backchannel

        debuggee.setup()
        backchannel.send(sys.argv)

    def expand(args):
        if expansion != "expand":
            return
        log.info("Before expansion: {0}", args)
        for i, arg in enumerate(args):
            if arg.startswith("$"):
                args[i] = arg[1:]
        log.info("After expansion: {0}", args)

    class Session(debug.Session):
        def run_in_terminal(self, args, cwd, env):
            expand(args)
            return super().run_in_terminal(args, cwd, env)

    argslist = ["0", "$1", "2"]
    args = argslist if expansion == "preserve" else " ".join(argslist)
    with Session() as session:
        backchannel = session.open_backchannel()
        with run(session, target(code_to_debug, args=args)):
            pass

        argv = backchannel.receive()

    expand(argslist)
    assert argv == [some.str] + argslist



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

import tests
from tests import debug
from tests.debug import runners
from tests.patterns import some


if not tests.full:

    @pytest.fixture(params=[runners.launch, runners.attach_connect["cli"]])
    def run(request):
        return request.param


def test_with_path_mappings(pyfile, tmpdir, target, run):
    @pyfile
    def code_to_debug():
        import debuggee

        debuggee.setup()

        def full_function():
            # Note that this function is not called, it's there just to make the mapping explicit.
            print("cell1 line 2")  # @map_to_cell1_line_2
            print("cell1 line 3")  # @map_to_cell1_line_3

            print("cell2 line 2")  # @map_to_cell2_line_2
            print("cell2 line 3")  # @map_to_cell2_line_3

        def strip_lines(s):
            return "\n".join([line.strip() for line in s.splitlines()])

        def create_code():
            cell1_code = compile(
                strip_lines(
                    """ # line 1
            a = 1  # line 2
            b = 2  # line 3
            """
                ),
                "<cell1>",
                "exec",
            )

            cell2_code = compile(
                strip_lines(
                    """# line 1
            c = 3  # line 2
            d = 4  # line 3
            """
                ),
                "<cell2>",
                "exec",
            )

            return {"cell1": cell1_code, "cell2": cell2_code}

        code = create_code()
        exec(code["cell1"], {})
        exec(code["cell1"], {})

        exec(code["cell2"], {})
        exec(code["cell2"], {})
        print("ok")

    with debug.Session() as session:
        map_to_cell_1_line2 = code_to_debug.lines["map_to_cell1_line_2"]
        map_to_cell_2_line2 = code_to_debug.lines["map_to_cell2_line_2"]

        source_entry = code_to_debug.strpath
        if sys.platform == "win32":
            # Check if it matches even not normalized.
            source_entry = source_entry[0].lower() + source_entry[1:].upper()
            source_entry = source_entry.replace("\\", "/")

        with run(session, target(code_to_debug)):
            # Set breakpoints first and the map afterwards to make sure that it's reapplied.
            session.set_breakpoints(code_to_debug, [map_to_cell_1_line2])

            session.request(
                "setPydevdSourceMap",
                {
                    "source": {"path": source_entry},
                    "pydevdSourceMaps": [
                        {
                            "line": map_to_cell_1_line2,
                            "endLine": map_to_cell_1_line2 + 1,
                            "runtimeSource": {"path": "<cell1>"},
                            "runtimeLine": 2,
                        },
                        {
                            "line": map_to_cell_2_line2,
                            "endLine": map_to_cell_2_line2 + 1,
                            "runtimeSource": {"path": "<cell2>"},
                            "runtimeLine": 2,
                        },
                    ],
                },
            )

        session.wait_for_stop(
            "breakpoint",
            expected_frames=[some.dap.frame(code_to_debug, line=map_to_cell_1_line2)],
        )

        session.set_breakpoints(code_to_debug, [map_to_cell_2_line2])
        # Leave only the cell2 mapping.
        session.request(
            "setPydevdSourceMap",
            {
                "source": {"path": source_entry},
                "pydevdSourceMaps": [
                    {
                        "line": map_to_cell_2_line2,
                        "endLine": map_to_cell_2_line2 + 1,
                        "runtimeSource": {"path": "<cell2>"},
                        "runtimeLine": 2,
                    }
                ],
            },
        )

        session.request_continue()
        session.wait_for_stop(
            "breakpoint",
            expected_frames=[some.dap.frame(code_to_debug, line=map_to_cell_2_line2)],
        )

        # Remove the cell2 mapping so that it doesn't stop again.
        session.request(
            "setPydevdSourceMap",
            {
                "source": {"path": source_entry},
                "pydevdSourceMaps": [
                    {
                        "line": map_to_cell_1_line2,
                        "endLine": map_to_cell_1_line2 + 1,
                        "runtimeSource": {"path": "<cell1>"},
                        "runtimeLine": 2,
                    }
                ],
            },
        )

        session.request_continue()



import unittest
import os
import sys
import pytest

# Note: ant.jar and junit.jar must be in the PYTHONPATH (see jython_test_deps)

IS_JYTHON = False
if sys.platform.find("java") != -1:
    IS_JYTHON = True
    from _pydev_bundle._pydev_jy_imports_tipper import ismethod
    from _pydev_bundle._pydev_jy_imports_tipper import isclass
    from _pydev_bundle._pydev_jy_imports_tipper import dir_obj
    from _pydev_bundle import _pydev_jy_imports_tipper
    from java.lang.reflect import Method  # @UnresolvedImport
    from java.lang import System  # @UnresolvedImport
    from java.lang import String  # @UnresolvedImport
    from java.lang.System import arraycopy  # @UnresolvedImport
    from java.lang.System import out  # @UnresolvedImport
    import java.lang.String  # @UnresolvedImport
    import org.python.core.PyDictionary  # @UnresolvedImport

__DBG = 0


def dbg(s):
    if __DBG:
        sys.stdout.write("%s\n" % (s,))


@pytest.mark.skipif(not IS_JYTHON, reason="Jython related test")
class TestMod(unittest.TestCase):
    def assert_args(self, tok, args, tips):
        for a in tips:
            if tok == a[0]:
                self.assertEqual(args, a[2])
                return
        raise AssertionError("%s not in %s", tok, tips)

    def assert_in(self, tok, tips):
        self.assertEqual(4, len(tips[0]))
        for a in tips:
            if tok == a[0]:
                return a
        s = ""
        for a in tips:
            s += str(a)
            s += "\n"
        raise AssertionError("%s not in %s" % (tok, s))

    def test_imports1a(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("java.util.HashMap")
        if f is None:
            return  # Not ok with java 9

        assert f.endswith("rt.jar")

    def test_imports1c(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("java.lang.Class")
        if f is None:
            return  # Not ok with java 9
        assert f.endswith("rt.jar")

    def test_imports1b(self):
        try:
            f, tip = _pydev_jy_imports_tipper.generate_tip("__builtin__.m")
            self.fail("err")
        except:
            pass

    def test_imports1(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("junit.framework.TestCase")
        assert f.endswith("junit.jar")
        ret = self.assert_in("assertEquals", tip)

    #        self.assertEqual('', ret[2])

    def test_imports2(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("junit.framework")
        assert f.endswith("junit.jar")
        ret = self.assert_in("TestCase", tip)
        self.assertEqual("", ret[2])

    def test_imports2a(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("org.apache.tools.ant")
        assert f.endswith("ant.jar")
        ret = self.assert_in("Task", tip)
        self.assertEqual("", ret[2])

    def test_imports3(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("os")
        assert f.endswith("os.py")
        ret = self.assert_in("path", tip)
        self.assertEqual("", ret[2])

    def test_tip_on_string(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("string")
        self.assert_in("join", tip)
        self.assert_in("uppercase", tip)

    def test_imports(self):
        tip = _pydev_jy_imports_tipper.generate_tip("__builtin__")[1]
        self.assert_in("tuple", tip)
        self.assert_in("RuntimeError", tip)
        self.assert_in("RuntimeWarning", tip)

    def test_imports5(self):
        f, tip = _pydev_jy_imports_tipper.generate_tip("java.lang")
        if f is None:
            return  # Not ok with java 9
        assert f.endswith("rt.jar")
        tup = self.assert_in("String", tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_CLASS), tup[3])

        tip = _pydev_jy_imports_tipper.generate_tip("java")[1]
        tup = self.assert_in("lang", tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_IMPORT), tup[3])

        tip = _pydev_jy_imports_tipper.generate_tip("java.lang.String")[1]
        tup = self.assert_in("indexOf", tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])

        tip = _pydev_jy_imports_tipper.generate_tip("java.lang.String")[1]
        tup = self.assert_in("charAt", tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])
        self.assertEqual("(int)", tup[2])

        tup = self.assert_in("format", tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])
        self.assertEqual("(string, objectArray)", tup[2])
        self.assertTrue(tup[1].find("[Ljava.lang.Object;") == -1)

        tup = self.assert_in("getBytes", tip)
        self.assertEqual(str(_pydev_jy_imports_tipper.TYPE_FUNCTION), tup[3])
        assert "[B" not in tup[1]
        assert "byte[]" in tup[1]

        f, tip = _pydev_jy_imports_tipper.generate_tip("__builtin__.str")
        assert f is None or f.endswith("jython.jar")  # Depends on jython version
        self.assert_in("find", tip)

        f, tip = _pydev_jy_imports_tipper.generate_tip("__builtin__.dict")
        assert f is None or f.endswith("jython.jar")  # Depends on jython version
        self.assert_in("get", tip)


@pytest.mark.skipif(not IS_JYTHON, reason="Jython related test")
class TestSearch(unittest.TestCase):
    def test_search_on_jython(self):
        assert _pydev_jy_imports_tipper.search_definition("os")[0][0].split(os.sep)[-1] in ("javaos.py", "os.py")
        self.assertEqual(0, _pydev_jy_imports_tipper.search_definition("os")[0][1])

        assert _pydev_jy_imports_tipper.search_definition("os.makedirs")[0][0].split(os.sep)[-1] in ("javaos.py", "os.py")
        self.assertNotEqual(0, _pydev_jy_imports_tipper.search_definition("os.makedirs")[0][1])

        # print _pydev_jy_imports_tipper.search_definition('os.makedirs')


@pytest.mark.skipif(not IS_JYTHON, reason="Jython related test")
class TestCompl(unittest.TestCase):
    def test_getting_info_on_jython(self):
        dbg("\n\n--------------------------- java")
        assert not ismethod(java)[0]
        assert not isclass(java)
        assert _pydev_jy_imports_tipper.ismodule(java)

        dbg("\n\n--------------------------- java.lang")
        assert not ismethod(java.lang)[0]
        assert not isclass(java.lang)
        assert _pydev_jy_imports_tipper.ismodule(java.lang)

        dbg("\n\n--------------------------- Method")
        assert not ismethod(Method)[0]
        assert isclass(Method)

        dbg("\n\n--------------------------- System")
        assert not ismethod(System)[0]
        assert isclass(System)

        dbg("\n\n--------------------------- String")
        assert not ismethod(System)[0]
        assert isclass(String)
        assert len(dir_obj(String)) > 10

        dbg("\n\n--------------------------- arraycopy")
        isMet = ismethod(arraycopy)
        assert isMet[0]
        assert (
            isMet[1][0].basic_as_str()
            == "function:arraycopy args=['java.lang.Object', 'int', 'java.lang.Object', 'int', 'int'], varargs=None, kwargs=None, docs:None"
        )
        assert not isclass(arraycopy)

        dbg("\n\n--------------------------- out")
        isMet = ismethod(out)
        assert not isMet[0]
        assert not isclass(out)

        dbg("\n\n--------------------------- out.println")
        isMet = ismethod(out.println)  # @UndefinedVariable
        assert isMet[0]
        assert len(isMet[1]) == 10
        self.assertEqual(isMet[1][0].basic_as_str(), "function:println args=[], varargs=None, kwargs=None, docs:None")
        assert isMet[1][1].basic_as_str() == "function:println args=['long'], varargs=None, kwargs=None, docs:None"
        assert not isclass(out.println)  # @UndefinedVariable

        dbg("\n\n--------------------------- str")
        isMet = ismethod(str)
        # the code below should work, but is failing on jython 22a1
        # assert isMet[0]
        # assert isMet[1][0].basic_as_str() == "function:str args=['org.python.core.PyObject'], varargs=None, kwargs=None, docs:None"
        assert not isclass(str)

        def met1():
            a = 3
            return a

        dbg("\n\n--------------------------- met1")
        isMet = ismethod(met1)
        assert isMet[0]
        assert isMet[1][0].basic_as_str() == "function:met1 args=[], varargs=None, kwargs=None, docs:None"
        assert not isclass(met1)

        def met2(arg1, arg2, *vararg, **kwarg):
            """docmet2"""

            a = 1
            return a

        dbg("\n\n--------------------------- met2")
        isMet = ismethod(met2)
        assert isMet[0]
        assert isMet[1][0].basic_as_str() == "function:met2 args=['arg1', 'arg2'], varargs=vararg, kwargs=kwarg, docs:docmet2"
        assert not isclass(met2)


# Run for jython in command line:

# On Windows:
# c:/bin/jython2.7.0/bin/jython.exe -Dpython.path=jython_test_deps/ant.jar;jython_test_deps/junit.jar -m py.test tests/test_jysimpleTipper.py

# On Linux (different path separator for jars)
# jython -Dpython.path=jython_test_deps/ant.jar:jython_test_deps/junit.jar -m py.test tests/test_jysimpleTipper.py



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import pytest
import sys

from debugpy.common import sockets


class TestSocketServerReuse(object):
    HOST1 = "127.0.0.1"
    # NOTE: Windows allows loopback range 127/8. Some flavors of Linux support
    # 127/8 range. Mac by default supports only 127/0. Configuring /etc/network/interface
    # for this one test is overkill so use '0.0.0.0' on Mac instead.
    HOST2 = "127.0.0.2" if sys.platform != "darwin" else "0.0.0.0"

    def test_reuse_same_address_port(self):
        # NOTE: This test should ensure that same address port can be used by two
        # sockets. This to prevent accidental changes to socket options. In Windows
        # SO_REUSEADDR flag allows two sockets to bind to same address:port combination.
        # Windows should always use SO_EXCLUSIVEADDRUSE
        sock1 = sockets.create_server(self.HOST1, 0)
        try:
            _, PORT1 = sock1.getsockname()
            with pytest.raises(Exception):
                sockets.create_server(self.HOST1, PORT1)
        finally:
            sockets.close_socket(sock1)

    def test_reuse_same_port(self):
        try:
            sock1, sock2 = None, None
            sock1 = sockets.create_server(self.HOST1, 0)
            _, PORT1 = sock1.getsockname()
            sock2 = sockets.create_server(self.HOST2, PORT1)
            assert sock1.getsockname() == (self.HOST1, PORT1)
            assert sock2.getsockname() == (self.HOST2, PORT1)
        except Exception:
            pytest.fail()
        finally:
            if sock1 is not None:
                sockets.close_socket(sock1)
            if sock2 is not None:
                sockets.close_socket(sock2)



from contextlib import contextmanager

import pytest

from _pydev_bundle.pydev_override import overrides
from tests_python.debugger_fixtures import DebuggerRunnerSimple, debugger_runner_simple
from tests_python.debugger_unittest import (
    AbstractWriterThread,
    SHOW_OTHER_DEBUG_INFO,
    start_in_daemon_thread,
    wait_for_condition,
    IS_JYTHON,
)
from _pydev_bundle.pydev_localhost import get_socket_names, get_socket_name
from _pydev_bundle.pydev_imports import xmlrpclib
from _pydev_bundle.pydev_imports import _queue as queue
from _pydev_bundle.pydev_imports import SimpleXMLRPCServer
import time
import socket

builtin_qualifier = "builtins"


@pytest.fixture
def console_setup(tmpdir):
    server_queue = queue.Queue()

    def notify_finished(more):
        server_queue.put(("notify_finished", more))
        return ""

    class ConsoleRunner(DebuggerRunnerSimple):
        @overrides(DebuggerRunnerSimple.add_command_line_args)
        def add_command_line_args(self, args, dap=False):
            port, client_port = get_socket_names(2, close=True)
            args.extend((writer.get_pydevconsole_file(), str(port[1]), str(client_port[1])))
            self.port = port
            self.client_port = client_port

            server = SimpleXMLRPCServer(client_port)
            server.register_function(notify_finished, "NotifyFinished")
            start_in_daemon_thread(server.serve_forever, [])

            self.proxy = xmlrpclib.ServerProxy("http://%s:%s/" % port)

            return args

    class WriterThread(AbstractWriterThread):
        if IS_JYTHON:
            EXPECTED_RETURNCODE = "any"

        @overrides(AbstractWriterThread.additional_output_checks)
        def additional_output_checks(self, stdout, stderr):
            print("output found: %s - %s" % (stdout, stderr))

        @overrides(AbstractWriterThread.write_dump_threads)
        def write_dump_threads(self):
            pass  # no-op (may be called on timeout).

        def execute_line(self, command, more=False):
            runner.proxy.execLine(command)
            assert server_queue.get(timeout=5.0) == ("notify_finished", more)

        def hello(self):
            def _hello():
                try:
                    msg = runner.proxy.hello("ignored")
                    if msg is not None:
                        if isinstance(msg, (list, tuple)):
                            msg = next(iter(msg))
                        if msg.lower().startswith("hello"):
                            return True
                except:
                    # That's ok, communication still not ready.
                    pass

                return False

            wait_for_condition(_hello)

        def close(self):
            try:
                runner.proxy.close()
            except:
                # Ignore any errors on close.
                pass

        def connect_to_debugger(self, debugger_port):
            runner.proxy.connectToDebugger(debugger_port)

    runner = ConsoleRunner(tmpdir)
    writer = WriterThread()

    class CaseSetup(object):
        @contextmanager
        def check_console(self, **kwargs):
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            self.writer = writer

            args = runner.get_command_line()

            args = runner.add_command_line_args(args)

            if SHOW_OTHER_DEBUG_INFO:
                print("executing: %s" % (" ".join(args),))
            try:
                with runner.run_process(args, writer) as dct_with_stdout_stder:
                    writer.get_stdout = lambda: "".join(dct_with_stdout_stder["stdout"])
                    writer.get_stderr = lambda: "".join(dct_with_stdout_stder["stderr"])

                    # Make sure communication is setup.
                    writer.hello()
                    yield writer
            finally:
                writer.log = []

            stdout = dct_with_stdout_stder["stdout"]
            stderr = dct_with_stdout_stder["stderr"]
            writer.additional_output_checks("".join(stdout), "".join(stderr))

    return CaseSetup()


def test_console_simple(console_setup):
    with console_setup.check_console() as writer:
        writer.execute_line("a = 10")
        writer.execute_line('print("TEST SUCEEDED")')
        writer.close()
        writer.finished_ok = True


def test_console_debugger_connected(console_setup):
    class _DebuggerWriterThread(AbstractWriterThread):
        FORCE_KILL_PROCESS_WHEN_FINISHED_OK = True

        def __init__(self):
            AbstractWriterThread.__init__(self)
            socket_name = get_socket_name(close=True)
            server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(socket_name)
            server_socket.listen(1)
            self.port = socket_name[1]
            self.__server_socket = server_socket

        def run(self):
            print("waiting for second process")
            self.sock, addr = self.__server_socket.accept()
            print("accepted second process")

            from tests_python.debugger_unittest import ReaderThread

            self.reader_thread = ReaderThread(self.sock)
            self.reader_thread.start()

            self._sequence = -1
            # initial command is always the version
            self.write_version()
            self.log.append("start_socket")
            self.write_make_initial_run()
            time.sleep(1)

            seq = self.write_list_threads()
            msg = self.wait_for_list_threads(seq)
            assert msg.thread["name"] == "MainThread"
            assert msg.thread["id"] == "console_main"

            self.write_get_frame("console_main", "1")
            self.wait_for_vars(
                [
                    [
                        '<var name="a" type="int" qualifier="%s" value="int: 10"' % (builtin_qualifier,),
                        '<var name="a" type="int"  value="int',  # jython
                    ],
                ]
            )

            self.finished_ok = True

    with console_setup.check_console() as writer:
        writer.execute_line("a = 10")

        debugger_writer_thread = _DebuggerWriterThread()
        debugger_writer_thread.start()
        writer.connect_to_debugger(debugger_writer_thread.port)

        wait_for_condition(lambda: debugger_writer_thread.finished_ok)
        writer.execute_line('print("TEST SUCEEDED")')

        writer.close()
        writer.finished_ok = True



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import pickle
import pytest
import subprocess
import sys

# This is used for mocking environment variables
# See https://docs.python.org/3/library/unittest.mock-examples.html for more info
from unittest import mock

from debugpy.common import log
from tests.patterns import some


@pytest.fixture
def cli(pyfile):
    @pyfile
    def cli_parser():
        import os
        import pickle
        import sys
        from debugpy.server import cli

        try:
            cli.parse_args()
        except Exception as exc:
            os.write(1, pickle.dumps(exc))
            sys.exit(1)

        # Check that sys.argv has the correct type after parsing - there should be no bytes.
        assert all(isinstance(s, str) for s in sys.argv)

        # We only care about options that correspond to public switches.
        options = {
            name: getattr(cli.options, name)
            for name in [
                "address",
                "config",
                "log_to",
                "log_to_stderr",
                "mode",
                "target",
                "target_kind",
                "wait_for_client",
            ]
        }

        # Serialize the command line args and the options to stdout
        os.write(1, pickle.dumps([sys.argv[1:], options]))

    def parse(args):
        log.debug("Parsing argv: {0!r}", args)
        try:
            # Run the CLI parser in a subprocess, and capture its output.
            output = subprocess.check_output(
                [sys.executable, "-u", cli_parser.strpath] + args
            )

            # Deserialize the output and return the parsed argv and options.
            argv, options = pickle.loads(output)
        except subprocess.CalledProcessError as exc:
            raise pickle.loads(exc.output)

        log.debug("Adjusted sys.argv: {0!r}", argv)
        log.debug("Parsed options: {0!r}", options)
        return argv, options

    return parse


# Test a combination of command line switches
@pytest.mark.parametrize("target_kind", ["file", "module", "code"])
@pytest.mark.parametrize("mode", ["listen", "connect"])
@pytest.mark.parametrize("address", ["8888", "localhost:8888"])
@pytest.mark.parametrize("wait_for_client", ["", "wait_for_client"])
@pytest.mark.parametrize("script_args", ["", "script_args"])
def test_targets(cli, target_kind, mode, address, wait_for_client, script_args):
    expected_options = {
        "mode": mode,
        "target_kind": target_kind,
        "wait_for_client": False
    }

    args = ["--" + mode, address]

    host, sep, port = address.partition(":")
    if sep:
        expected_options["address"] = (host, int(port))
    else:
        expected_options["address"] = ("127.0.0.1", int(address))

    if wait_for_client:
        args += ["--wait-for-client"]
        expected_options["wait_for_client"] = True

    if target_kind == "file":
        target = "spam.py"
        args += [target]
    elif target_kind == "module":
        target = "spam"
        args += ["-m", target]
    elif target_kind == "code":
        target = "123"
        args += ["-c", target]
    else:
        pytest.fail(target_kind)
    expected_options["target"] = target

    if script_args:
        script_args = [
            "ham",
            "--listen",
            "--wait-for-client",
            "-y",
            "spam",
            "--",
            "--connect",
            "-c",
            "--something",
            "-m",
        ]
        args += script_args
    else:
        script_args = []

    argv, options = cli(args)
    assert argv == script_args
    assert options == some.dict.containing(expected_options)

@pytest.mark.parametrize("value", [True, False])
def test_configure_subProcess(cli, value):
    args = ["--listen", "8888", "--configure-subProcess", str(value), "spam.py"]
    _, options = cli(args)

    assert options["config"]["subProcess"] == value

@pytest.mark.parametrize("value", [True, False])
def test_configure_subProcess_from_environment(cli, value):
    args = ["--listen", "8888", "spam.py"]
    with mock.patch.dict(os.environ, {"DEBUGPY_EXTRA_ARGV": "--configure-subProcess " + str(value)}):
        _, options = cli(args)

        assert options["config"]["subProcess"] == value

def test_unsupported_switch(cli):
    with pytest.raises(ValueError) as ex:
        cli(["--listen", "8888", "--xyz", "123", "spam.py"])
    
    assert "unrecognized switch --xyz" in str(ex.value)

def test_unsupported_switch_from_environment(cli):
    with pytest.raises(ValueError) as ex:
        with mock.patch.dict(os.environ, {"DEBUGPY_EXTRA_ARGV": "--xyz 123"}):
            cli(["--listen", "8888", "spam.py"])
    
    assert "unrecognized switch --xyz" in str(ex.value)

def test_unsupported_configure(cli):
    with pytest.raises(ValueError) as ex:
        cli(["--connect", "127.0.0.1:8888", "--configure-xyz", "123", "spam.py"])
    
    assert "unknown property 'xyz'" in str(ex.value)

def test_unsupported_configure_from_environment(cli):
    with pytest.raises(ValueError) as ex:
        with mock.patch.dict(os.environ, {"DEBUGPY_EXTRA_ARGV": "--configure-xyz 123"}):
            cli(["--connect", "127.0.0.1:8888", "spam.py"])

    assert "unknown property 'xyz'" in str(ex.value)

def test_address_required(cli):
    with pytest.raises(ValueError) as ex:
        cli(["-m", "spam"])
    
    assert "either --listen or --connect is required" in str(ex.value)

def test_missing_target(cli):
    with pytest.raises(ValueError) as ex:
        cli(["--listen", "8888"])
    
    assert "missing target" in str(ex.value)

def test_duplicate_switch(cli):
    with pytest.raises(ValueError) as ex:
        cli(["--listen", "8888", "--listen", "9999", "spam.py"])
    
    assert "duplicate switch on command line: --listen" in str(ex.value)

def test_duplicate_switch_from_environment(cli):
    with pytest.raises(ValueError) as ex:
        with mock.patch.dict(os.environ, {"DEBUGPY_EXTRA_ARGV": "--listen 8888 --listen 9999"}):
            cli(["spam.py"])
    
    assert "duplicate switch from environment: --listen" in str(ex.value)

# Test that switches can be read from the environment
def test_read_switches_from_environment(cli):
    args = ["spam.py"]

    with mock.patch.dict(os.environ, {"DEBUGPY_EXTRA_ARGV": "--connect 5678"}):
        _, options = cli(args)

        assert options["mode"] == "connect"
        assert options["address"] == ("127.0.0.1", 5678)
        assert options["target"] == "spam.py"

# Test that command line switches override environment variables
def test_override_environment_switch(cli):
    args = ["--connect", "8888", "spam.py"]

    with mock.patch.dict(os.environ, {"DEBUGPY_EXTRA_ARGV": "--connect 5678"}):
        _, options = cli(args)

        assert options["mode"] == "connect"
        assert options["address"] == ("127.0.0.1", 8888)
        assert options["target"] == "spam.py"

# Test that script args (passed to target) are preserved
def test_script_args(cli):
    args = ["--listen", "8888", "spam.py", "arg1", "arg2"]
    argv, options = cli(args)

    assert argv == ["arg1", "arg2"]
    assert options["target"] == "spam.py"


# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""Tests for product code that is shared between the adapter and the server.
"""



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import io
import os
import shutil
import sys

from debugpy.common import json, log


def write_title(title, stream=None, sep="~"):
    """Write a section title.
    If *stream* is None sys.stderr will be used, *sep* is used to
    draw the line.
    """
    if stream is None:
        stream = sys.stderr
    width, height = shutil.get_terminal_size()
    fill = int((width - len(title) - 2) / 2)
    line = " ".join([sep * fill, title, sep * fill])
    if len(line) < width:
        line += sep * (width - len(line))
    stream.write("\n" + line + "\n")

def dump():
    if log.log_dir is None:
        return
    log.info("Dumping logs from {0}", json.repr(log.log_dir))

    for dirpath, dirnames, filenames in os.walk(log.log_dir):
        for name in sorted(filenames):
            if not name.startswith("debugpy") and not name.startswith("pydevd"):
                continue
            try:
                path = os.path.join(dirpath, name)
                with io.open(path, encoding="utf-8", errors="backslashreplace") as f:
                    s = f.read()
            except Exception:
                pass
            else:
                path = os.path.relpath(path, log.log_dir)
                write_title(path)
                print(s, file=sys.stderr)
